#!/usr/bin/env python
"""测试单个worker重启后自动恢复pending消息的机制"""

import asyncio
import json
import logging
import os
import sys
import time
from typing import Dict, List, Optional
from datetime import datetime, timezone

import redis.asyncio as redis
from redis.asyncio import Redis
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from sqlalchemy.orm import sessionmaker
from sqlalchemy import text

from jettask.webui.config import PostgreSQLConfig, RedisConfig
from jettask.utils.serializer import dumps_str

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)

class SelfRecoveryTester:
    """测试单个worker自我恢复机制"""
    
    def __init__(self):
        self.pg_config = PostgreSQLConfig(
            host=os.getenv('JETTASK_PG_HOST', 'localhost'),
            port=int(os.getenv('JETTASK_PG_PORT', '5432')),
            database=os.getenv('JETTASK_PG_DB', 'jettask'),
            user=os.getenv('JETTASK_PG_USER', 'jettask'),
            password=os.getenv('JETTASK_PG_PASSWORD', '123456'),
        )
        
        self.redis_config = RedisConfig(
            host=os.getenv('REDIS_HOST', 'localhost'),
            port=int(os.getenv('REDIS_PORT', '6379')),
            db=int(os.getenv('REDIS_DB', '0')),
            password=os.getenv('REDIS_PASSWORD'),
        )
        
        self.prefix = "jettask"
        self.redis_client: Optional[Redis] = None
        
    async def setup(self):
        """初始化连接"""
        self.redis_client = await redis.Redis(
            host=self.redis_config.host,
            port=self.redis_config.port,
            db=self.redis_config.db,
            password=self.redis_config.password,
            decode_responses=False
        )
        
    async def cleanup(self):
        """清理连接"""
        if self.redis_client:
            await self.redis_client.aclose()
    
    async def test_queue_self_recovery(self):
        """测试队列消息的自我恢复机制"""
        logger.info("=" * 60)
        logger.info("测试单个worker重启后自动恢复自己的pending消息")
        logger.info("=" * 60)
        
        queue_name = 'SELF_RECOVERY_QUEUE'
        stream_key = f"{self.prefix}:QUEUE:{queue_name}"
        consumer_group = f"{self.prefix}_pg_consumer1"
        consumer_name = "self_recovery_worker"
        
        # 1. 清理旧数据
        logger.info("\n1. 清理旧数据...")
        try:
            await self.redis_client.delete(stream_key)
        except:
            pass
        
        # 2. 发送测试消息
        logger.info("\n2. 发送测试消息到队列...")
        message_ids = []
        for i in range(5):
            task_data = {
                'name': f'self_recovery_task_{i}',
                'queue': queue_name,
                'priority': 1,
                'trigger_time': time.time(),
                'test_id': f'self_recovery_{i}'
            }
            
            msg_id = await self.redis_client.xadd(
                stream_key,
                {'data': dumps_str(task_data)}
            )
            message_ids.append(msg_id)
            logger.info(f"  - 发送消息 {i}: {msg_id}")
        
        # 3. 创建消费者组
        logger.info("\n3. 创建消费者组...")
        try:
            await self.redis_client.xgroup_create(
                stream_key, consumer_group, id='0', mkstream=True
            )
            logger.info(f"  - 创建消费者组: {consumer_group}")
        except:
            pass
        
        # 4. 模拟worker第一次读取消息（读取新消息）
        logger.info(f"\n4. 模拟 {consumer_name} 第一次读取新消息但不ACK（模拟突然挂掉）...")
        
        messages = await self.redis_client.xreadgroup(
            consumer_group,
            consumer_name,
            {stream_key: '>'},  # '>' 表示读取新消息
            count=5,
            block=1000
        )
        
        first_read_ids = []
        if messages:
            logger.info(f"  - {consumer_name} 读取了 {len(messages[0][1])} 条新消息")
            for msg_id, data in messages[0][1]:
                first_read_ids.append(msg_id)
                logger.info(f"    - 消息ID: {msg_id}")
        
        # 5. 检查pending消息
        logger.info("\n5. 检查pending消息状态...")
        pending_info = await self.redis_client.xpending(stream_key, consumer_group)
        logger.info(f"  - 总pending消息数: {pending_info['pending']}")
        
        if pending_info['pending'] > 0:
            detailed = await self.redis_client.xpending_range(
                stream_key, consumer_group,
                min='-', max='+', count=10
            )
            for msg in detailed:
                logger.info(f"    - 消息 {msg['message_id']}: consumer={msg['consumer']}, times_delivered={msg['times_delivered']}")
        
        # 6. 模拟同一个worker重启后读取pending消息（不使用XCLAIM）
        logger.info(f"\n6. 模拟 {consumer_name} 重启后读取自己的pending消息...")
        logger.info("  - 使用 '0' 作为起始ID，这会让consumer读取自己的pending消息")
        
        # 先读取pending消息（使用0作为起始ID）
        pending_messages = await self.redis_client.xreadgroup(
            consumer_group,
            consumer_name,  # 同一个consumer name
            {stream_key: '0'},  # '0' 或 '0-0' 表示从头读取pending消息
            count=10,
            block=0  # 不阻塞
        )
        
        recovered_ids = []
        if pending_messages:
            logger.info(f"  - {consumer_name} 成功读取到 {len(pending_messages[0][1])} 条自己的pending消息")
            for msg_id, data in pending_messages[0][1]:
                recovered_ids.append(msg_id)
                logger.info(f"    - 恢复消息ID: {msg_id}")
                
            # ACK这些消息
            if recovered_ids:
                await self.redis_client.xack(stream_key, consumer_group, *recovered_ids)
                logger.info(f"  - {consumer_name} ACK了 {len(recovered_ids)} 条恢复的消息")
        else:
            logger.info(f"  - {consumer_name} 没有读取到pending消息")
        
        # 7. 再次检查pending消息
        logger.info("\n7. 再次检查pending消息状态...")
        pending_info_after = await self.redis_client.xpending(stream_key, consumer_group)
        logger.info(f"  - 总pending消息数: {pending_info_after['pending']}")
        
        # 8. 验证结果
        logger.info("\n" + "=" * 60)
        if pending_info['pending'] > 0 and pending_info_after['pending'] == 0:
            logger.info("✓ 单个worker自我恢复测试通过！")
            logger.info("  worker重启后成功读取并处理了自己的pending消息，无需XCLAIM")
            return True
        else:
            logger.warning(f"✗ 单个worker自我恢复测试失败。")
            logger.warning(f"  恢复前pending: {pending_info['pending']}，恢复后pending: {pending_info_after['pending']}")
            return False
    
    async def test_task_changes_self_recovery(self):
        """测试TASK_CHANGES的自我恢复机制"""
        logger.info("\n" + "=" * 60)
        logger.info("测试TASK_CHANGES单个worker自我恢复机制")
        logger.info("=" * 60)
        
        change_stream_key = f"{self.prefix}:TASK_CHANGES"
        consumer_group = f"{self.prefix}_changes_consumer"
        consumer_name = "changes_self_recovery_worker"
        
        # 1. 清理旧数据
        logger.info("\n1. 清理旧数据...")
        try:
            await self.redis_client.delete(change_stream_key)
        except:
            pass
        
        # 2. 发送任务变更事件
        logger.info("\n2. 发送任务变更事件...")
        event_ids = []
        for i in range(5):
            event_data = {
                'event_id': f'self_recovery_change_{i}',
                'event_type': 'task_updated',
                'timestamp': str(time.time())
            }
            
            msg_id = await self.redis_client.xadd(
                change_stream_key,
                event_data
            )
            event_ids.append(msg_id)
            logger.info(f"  - 发送事件 {i}: {msg_id}")
        
        # 3. 创建消费者组
        logger.info("\n3. 创建消费者组...")
        try:
            await self.redis_client.xgroup_create(
                change_stream_key, consumer_group, id='0', mkstream=True
            )
            logger.info(f"  - 创建消费者组: {consumer_group}")
        except:
            pass
        
        # 4. 模拟worker第一次读取消息但不ACK
        logger.info(f"\n4. 模拟 {consumer_name} 第一次读取事件但不ACK...")
        
        messages = await self.redis_client.xreadgroup(
            consumer_group,
            consumer_name,
            {change_stream_key: '>'},
            count=5,
            block=1000
        )
        
        if messages:
            logger.info(f"  - {consumer_name} 读取了 {len(messages[0][1])} 条事件")
        
        # 5. 检查pending事件
        logger.info("\n5. 检查pending事件...")
        pending_info = await self.redis_client.xpending(change_stream_key, consumer_group)
        logger.info(f"  - 总pending事件数: {pending_info['pending']}")
        
        # 6. 模拟同一个worker重启后恢复
        logger.info(f"\n6. 模拟 {consumer_name} 重启后读取自己的pending事件...")
        
        # 读取pending消息
        pending_messages = await self.redis_client.xreadgroup(
            consumer_group,
            consumer_name,
            {change_stream_key: '0'},  # 从头读取pending
            count=10,
            block=0
        )
        
        recovered_ids = []
        if pending_messages:
            logger.info(f"  - {consumer_name} 成功恢复 {len(pending_messages[0][1])} 条pending事件")
            for msg_id, data in pending_messages[0][1]:
                recovered_ids.append(msg_id)
                
            # ACK消息
            if recovered_ids:
                await self.redis_client.xack(change_stream_key, consumer_group, *recovered_ids)
                logger.info(f"  - {consumer_name} ACK了 {len(recovered_ids)} 条恢复的事件")
        
        # 7. 再次检查pending事件
        logger.info("\n7. 再次检查pending事件...")
        pending_info_after = await self.redis_client.xpending(change_stream_key, consumer_group)
        logger.info(f"  - 总pending事件数: {pending_info_after['pending']}")
        
        # 8. 验证结果
        logger.info("\n" + "=" * 60)
        if pending_info['pending'] > 0 and pending_info_after['pending'] == 0:
            logger.info("✓ TASK_CHANGES自我恢复测试通过！")
            return True
        else:
            logger.warning(f"✗ TASK_CHANGES自我恢复测试失败")
            return False
    
    async def demonstrate_pg_consumer_recovery_flow(self):
        """演示pg_consumer实际的恢复流程"""
        logger.info("\n" + "=" * 60)
        logger.info("演示pg_consumer的实际恢复流程")
        logger.info("=" * 60)
        
        queue_name = 'PG_CONSUMER_DEMO'
        stream_key = f"{self.prefix}:QUEUE:{queue_name}"
        consumer_group = f"{self.prefix}_pg_consumer1"
        consumer_name = "pg_consumer_worker_1"
        
        # 清理
        try:
            await self.redis_client.delete(stream_key)
        except:
            pass
        
        # 发送消息
        logger.info("\n1. 发送5条消息...")
        for i in range(5):
            await self.redis_client.xadd(
                stream_key,
                {'data': dumps_str({'name': f'demo_task_{i}'})}
            )
        
        # 创建消费者组
        try:
            await self.redis_client.xgroup_create(
                stream_key, consumer_group, id='0', mkstream=True
            )
        except:
            pass
        
        # 模拟pg_consumer的消费逻辑
        logger.info(f"\n2. {consumer_name} 开始消费...")
        
        check_backlog = True
        lastid = "0-0"
        processed_count = 0
        
        while True:
            # 这是pg_consumer的实际逻辑
            myid = lastid if check_backlog else ">"
            
            logger.info(f"  - 读取消息，myid={myid}, check_backlog={check_backlog}")
            
            messages = await self.redis_client.xreadgroup(
                consumer_group,
                consumer_name,
                {stream_key: myid},
                count=2,  # 批量读取
                block=0 if check_backlog else 1000
            )
            
            if not messages or len(messages[0][1]) == 0:
                if check_backlog:
                    logger.info("  - 没有更多backlog消息，切换到读取新消息模式")
                    check_backlog = False
                    continue
                else:
                    logger.info("  - 没有新消息")
                    break
            
            # 处理消息
            msg_count = len(messages[0][1])
            logger.info(f"  - 读取到 {msg_count} 条消息")
            
            # 模拟处理失败（第一次只处理2条就"挂掉"）
            if processed_count < 2:
                processed_count += msg_count
                if processed_count >= 2:
                    logger.info("  - 模拟worker挂掉，不ACK剩余消息")
                    break
            
            # 正常ACK
            msg_ids = [msg[0] for msg in messages[0][1]]
            await self.redis_client.xack(stream_key, consumer_group, *msg_ids)
            logger.info(f"  - ACK了 {len(msg_ids)} 条消息")
            
            # 更新lastid
            lastid = messages[0][1][-1][0].decode('utf-8') if isinstance(messages[0][1][-1][0], bytes) else messages[0][1][-1][0]
            
            # 检查是否还有更多backlog
            check_backlog = msg_count >= 2
        
        # 检查pending
        pending_info = await self.redis_client.xpending(stream_key, consumer_group)
        logger.info(f"\n3. Worker挂掉后，pending消息数: {pending_info['pending']}")
        
        # 模拟重启后恢复
        logger.info(f"\n4. {consumer_name} 重启，继续消费...")
        
        check_backlog = True
        lastid = "0-0"  # 重新从0开始，会读取pending消息
        
        while True:
            myid = lastid if check_backlog else ">"
            
            logger.info(f"  - 读取消息，myid={myid}, check_backlog={check_backlog}")
            
            messages = await self.redis_client.xreadgroup(
                consumer_group,
                consumer_name,
                {stream_key: myid},
                count=10,
                block=0 if check_backlog else 1000
            )
            
            if not messages or len(messages[0][1]) == 0:
                if check_backlog:
                    logger.info("  - 没有更多pending/backlog消息")
                    check_backlog = False
                    continue
                else:
                    logger.info("  - 没有新消息，完成")
                    break
            
            msg_count = len(messages[0][1])
            logger.info(f"  - 恢复并处理 {msg_count} 条消息")
            
            # ACK消息
            msg_ids = [msg[0] for msg in messages[0][1]]
            await self.redis_client.xack(stream_key, consumer_group, *msg_ids)
            logger.info(f"  - ACK了 {len(msg_ids)} 条消息")
            
            # 更新lastid
            lastid = messages[0][1][-1][0].decode('utf-8') if isinstance(messages[0][1][-1][0], bytes) else messages[0][1][-1][0]
            check_backlog = msg_count >= 10
        
        # 最终检查
        pending_info_after = await self.redis_client.xpending(stream_key, consumer_group)
        logger.info(f"\n5. 恢复完成后，pending消息数: {pending_info_after['pending']}")
        
        logger.info("\n" + "=" * 60)
        logger.info("✓ 演示完成：pg_consumer使用check_backlog机制自动恢复pending消息")
        logger.info("  关键点：重启后设置lastid='0-0'和check_backlog=True")
        logger.info("  这样会先处理所有pending和未读消息，然后再读取新消息")


async def main():
    """主测试函数"""
    from dotenv import load_dotenv
    load_dotenv()
    
    tester = SelfRecoveryTester()
    await tester.setup()
    
    try:
        # 测试1：队列消息自我恢复
        queue_test = await tester.test_queue_self_recovery()
        await asyncio.sleep(1)
        
        # 测试2：TASK_CHANGES自我恢复
        changes_test = await tester.test_task_changes_self_recovery()
        await asyncio.sleep(1)
        
        # 演示pg_consumer的实际恢复流程
        await tester.demonstrate_pg_consumer_recovery_flow()
        
        # 总结
        logger.info("\n" + "=" * 60)
        logger.info("测试总结")
        logger.info("=" * 60)
        logger.info(f"队列消息自我恢复测试: {'✓ 通过' if queue_test else '✗ 失败'}")
        logger.info(f"TASK_CHANGES自我恢复测试: {'✓ 通过' if changes_test else '✗ 失败'}")
        
        if queue_test and changes_test:
            logger.info("\n🎉 所有测试通过！")
            logger.info("\n关键发现：")
            logger.info("1. 单个worker重启后，使用相同的consumer_name读取")
            logger.info("2. 设置起始ID为'0'或'0-0'，可以读取自己的pending消息")
            logger.info("3. 不需要XCLAIM，直接xreadgroup即可恢复")
            logger.info("4. pg_consumer的check_backlog机制完美支持这种恢复方式")
        
    except Exception as e:
        logger.error(f"测试出错: {e}", exc_info=True)
    finally:
        await tester.cleanup()


if __name__ == '__main__':
    asyncio.run(main())


#!/usr/bin/env python
"""启动任务监控服务"""

import uvicorn
import sys
import os

# 添加当前目录到Python路径
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

if __name__ == "__main__":
    print("Starting JetTask Monitor Backend...")
    print("API will be available at: http://localhost:8000")
    print("API docs: http://localhost:8000/docs")
    
    uvicorn.run(
        "backend.main:app",
        host="0.0.0.0",
        port=8000,
        reload=True,
        reload_dirs=[os.path.dirname(os.path.abspath(__file__))]
    )


import time
import traceback
from ..utils.serializer import dumps_str, loads_str
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple
from ..utils.traceback_filter import filter_framework_traceback

if TYPE_CHECKING:
    from ..core.app import Jettask
    from ..core.task import Task


class CommonExecutorMixin:
    """Mixin class containing common functionality for all executors"""
    
    app: "Jettask"
    last_refresh_pending_time: float
    pedding_count: int
    
    def parse_task_data(self, data: bytes) -> Tuple[str, str, str, list, dict]:
        """Parse task data from bytes"""
        data_str = data.decode("utf-8")
        parts = data_str.split("_____", 4)
        event_id = parts[0]
        task_name = parts[1]
        trigger_time = parts[2]
        args = loads_str(parts[3]) if parts[3] else []
        kwargs = loads_str(parts[4]) if parts[4] else {}
        return event_id, task_name, trigger_time, args, kwargs
    
    def get_task(self, task_name: str) -> Optional["Task"]:
        """Get task by name"""
        return self.app.get_task_by_name(task_name)
    
    def handle_task_error(self, event_id: str, task_name: str, error: Exception) -> Dict[str, Any]:
        """Handle task execution error"""
        # 使用过滤后的堆栈信息
        error_msg = filter_framework_traceback()
        result = {
            "status": "error",
            "task_name": task_name,
            "message": str(error),
            "traceback": error_msg
        }
        self.app.set_task_status(event_id, "error")
        self.app.set_data(event_id, dumps_str(result))
        return result
    
    def handle_task_success(self, event_id: str, task_name: str, result: Any) -> Dict[str, Any]:
        """Handle successful task execution"""
        result_data = {
            "status": "success",
            "task_name": task_name,
            "result": result
        }
        self.app.set_task_status(event_id, "success")
        self.app.set_data(event_id, dumps_str(result_data))
        return result_data
    
    def ack_message(self, data: bytes) -> None:
        """Acknowledge message processing"""
        self.app.ep.ack(data)
    
    def get_routing_data(self, kwargs: dict) -> Tuple[Optional[str], Optional[str]]:
        """Extract routing data from kwargs"""
        routing_key = kwargs.pop("routing_key", None)
        aggregation_key = kwargs.pop("aggregation_key", None)
        return routing_key, aggregation_key
    
    def should_process_routing(self, routing_key: Optional[str], aggregation_key: Optional[str]) -> bool:
        """Check if task should be processed based on routing"""
        if routing_key:
            # Check solo running state
            if self.app.is_solo_running_by_aggregation_key(aggregation_key):
                return False
            # Set solo running state
            self.app.set_solo_running_by_aggregation_key(aggregation_key)
        return True
    
    def clear_routing_state(self, routing_key: Optional[str], aggregation_key: Optional[str]) -> None:
        """Clear routing state after task completion"""
        if routing_key:
            self.app.clear_solo_running_by_aggregation_key(aggregation_key)
    
    def handle_urgent_retry(self, kwargs: dict, event_id: str, task_name: str, 
                          trigger_time: str, args: list) -> bool:
        """Handle urgent retry logic"""
        urgent_retry = kwargs.pop("urgent_retry", None)
        if urgent_retry:
            delay = kwargs.pop("delay", 0)
            self.app.delay(
                task_name=task_name,
                delay=delay,
                args=args,
                kwargs=kwargs,
                task_id=event_id,
                trigger_time=trigger_time,
                producer_type="urgent_retry",
            )
            self.app.set_task_status(event_id, "retry")
            return True
        return False
    
    def get_pending_count(self) -> int:
        """Get pending count with caching"""
        current_time = time.time()
        if current_time - self.last_refresh_pending_time > 1:
            self.pedding_count = self.app.ep.get_pending_count()
            self.last_refresh_pending_time = current_time
        return self.pedding_count
    
    def execute_task_lifecycle(self, task: "Task", event_id: str, trigger_time: str,
                             args: list, kwargs: dict, is_async: bool = False):
        """Execute task with lifecycle methods"""
        # This method will be implemented differently for sync/async executors
        raise NotImplementedError("Subclasses must implement execute_task_lifecycle")
    
    def format_batch_data(self, data: bytes) -> Tuple[str, str, list, dict]:
        """Parse batch task data"""
        data_str = data.decode("utf-8")
        parts = data_str.split("_____", 3)
        batch_id = parts[0]
        task_name = parts[1]
        args = loads_str(parts[2]) if parts[2] else []
        kwargs = loads_str(parts[3]) if parts[3] else {}
        return batch_id, task_name, args, kwargs
    
    def extract_batch_params(self, kwargs: dict) -> Tuple[list, list, str, Optional[str], Optional[str]]:
        """Extract batch-specific parameters from kwargs"""
        event_ids = kwargs.pop("event_ids", [])
        trigger_times = kwargs.pop("trigger_times", [])
        producer_type = kwargs.pop("producer_type", "normal")
        routing_key = kwargs.pop("routing_key", None)
        aggregation_key = kwargs.pop("aggregation_key", None)
        return event_ids, trigger_times, producer_type, routing_key, aggregation_key
    
    def ack_batch_events(self, event_ids: list) -> None:
        """Acknowledge multiple events in batch"""
        for event_id in event_ids:
            data = f"{event_id}_____batch_ack".encode()
            self.ack_message(data)
    
    def update_batch_status(self, event_ids: list, status: str, result: Any = None) -> None:
        """Update status for multiple events in batch"""
        result_str = dumps_str(result) if result is not None else None
        for event_id in event_ids:
            self.app.set_task_status(event_id, status)
            if result_str:
                self.app.set_data(event_id, result_str)


"""
简化的离线worker消息恢复模块
"""
import asyncio
import logging
from typing import Dict, List, Optional, Tuple
from redis.asyncio.lock import Lock as AsyncLock

import msgpack

logger = logging.getLogger(__name__)


class OfflineWorkerRecovery:
    """离线worker消息恢复处理器"""
    
    def __init__(self, async_redis_client, consumer_manager=None, redis_prefix='jettask', worker_prefix='WORKER', queue_formatter=None):
        self.async_redis_client = async_redis_client
        self.consumer_manager = consumer_manager
        self.redis_prefix = redis_prefix
        self.worker_prefix = worker_prefix
        self._stop_recovery = False
        # 队列格式化函数，默认使用 prefix:QUEUE:queue_name 格式
        self.queue_formatter = queue_formatter or (lambda q: f"{self.redis_prefix}:QUEUE:{q}")
        
    async def recover_offline_workers(self,
                                     queue: str,
                                     current_consumer_name: str = None,
                                     event_queue: Optional[asyncio.Queue] = None,
                                     process_message_callback: Optional[callable] = None,
                                     consumer_group_suffix: Optional[str] = None) -> int:
        """
        恢复指定队列的离线worker的pending消息
        """
        total_recovered = 0
        
        try:
            # 获取当前consumer名称
            if not current_consumer_name and self.consumer_manager:
                current_consumer_name = self.consumer_manager.get_consumer_name(queue)
            
            if not current_consumer_name:
                logger.error(f"Cannot get current consumer name for queue {queue}")
                return 0
                
            # logger.info(f"Starting recovery for queue {queue} with consumer {current_consumer_name}")
            
            # 获取所有离线worker
            offline_workers = await self._find_offline_workers(queue)
            # logger.info(f'{offline_workers=}')
            if not offline_workers:
                logger.debug(f"No offline workers found for queue {queue}")
                return 0
                
            logger.info(f"Found {len(offline_workers)} offline workers for queue {queue}")
            
            # 处理每个离线worker
            for worker_key, worker_data in offline_workers:
                if self._stop_recovery:
                    logger.info("Stopping recovery due to shutdown signal")
                    break
                
                # logger.info(f"Processing offline worker: {worker_key}")
                recovered = await self._recover_worker_messages(
                    queue=queue,
                    worker_key=worker_key,
                    worker_data=worker_data,
                    current_consumer_name=current_consumer_name,
                    event_queue=event_queue,
                    process_message_callback=process_message_callback,
                    consumer_group_suffix=consumer_group_suffix
                )
                
                total_recovered += recovered
                
        except Exception as e:
            logger.error(f"Error recovering offline workers for queue {queue}: {e}")
            
        return total_recovered
        
    async def _find_offline_workers(self, queue: str) -> List[Tuple[str, Dict]]:
        """查找指定队列的离线worker"""
        offline_workers = []
        
        try:
            # 扫描所有worker
            pattern = f"{self.redis_prefix}:{self.worker_prefix}:*"
            cursor = 0
            while True:
                cursor, keys = await self.async_redis_client.scan(
                    cursor=cursor,
                    match=pattern,
                    count=100
                )
                
                for worker_key in keys:
                    if isinstance(worker_key, bytes):
                        worker_key = worker_key.decode('utf-8')
                    
                    # 跳过非worker键
                    if any(x in worker_key for x in [':HISTORY:', ':REUSE:LOCK', ':REUSING']):
                        continue
                    
                    try:
                        worker_data = await self.async_redis_client.hgetall(worker_key)
                        if not worker_data:
                            continue
                        
                        # 解码二进制数据
                        decoded_worker_data = {}
                        for k, v in worker_data.items():
                            key = k.decode('utf-8') if isinstance(k, bytes) else k
                            value = v.decode('utf-8') if isinstance(v, bytes) else v
                            decoded_worker_data[key] = value
                        
                        # logger.info(f'{worker_key=} {decoded_worker_data=}')
                        # logger.info(f'{decoded_worker_data=}')
                        # 检查worker是否离线且消息未转移
                        is_alive = decoded_worker_data.get('is_alive', 'false').lower() == 'true'
                        messages_transferred = decoded_worker_data.get('messages_transferred', 'false').lower() == 'true'
                        # logger.info(f'{worker_key=} {is_alive=} {messages_transferred=} {not is_alive and not messages_transferred}')
                        # 找到离线且消息未转移的worker
                        if not is_alive and not messages_transferred:
                            queues_str = decoded_worker_data.get('queues', '')
                            worker_queues = queues_str.split(',') if queues_str else []
                            
                            # logger.info(f'{worker_queues=} {queue=}')
                            # 检查这个worker是否负责当前队列
                            if queue in worker_queues:
                                offline_workers.append((worker_key, decoded_worker_data))
                                
                    except Exception as e:
                        logger.error(f"Error processing worker key {worker_key}: {e}")
                        continue
                
                # 当cursor返回0时，表示扫描完成
                if cursor == 0:
                    break
                    
        except Exception as e:
            logger.error(f"Error finding offline workers: {e}")
            
        return offline_workers
        
    async def _recover_worker_messages(self,
                                      queue: str,
                                      worker_key: str,
                                      worker_data: Dict,
                                      current_consumer_name: str,
                                      event_queue: Optional[asyncio.Queue] = None,
                                      process_message_callback: Optional[callable] = None,
                                      consumer_group_suffix: Optional[str] = None) -> int:
        """
        恢复单个worker的pending消息
        
        简化逻辑：
        1. 使用离线worker的consumer_id构建consumer名称：{consumer_id}-{queue}
        2. 默认group名称就是{prefixed_queue}
        3. 直接获取并转移该consumer的pending消息
        """
        total_claimed = 0
        
        try:
            # worker_data 现在已经是解码后的字典
            consumer_id = worker_data.get('consumer_id')
            
            # 构建离线worker的consumer名称
            offline_consumer = f"{consumer_id}-{queue}"
            
            logger.info(f"Recovering messages from offline worker: {offline_consumer}")
            
            # 检查是否是同一个worker（避免自己接管自己的消息）
            if current_consumer_name == offline_consumer or current_consumer_name.startswith(f"{offline_consumer}:"):
                logger.info(f"Skipping {offline_consumer} - same worker")
                return 0
                
            # 使用分布式锁防止并发处理
            lock_key = f"{self.redis_prefix}:CLAIM:LOCK:{offline_consumer}"
            lock = AsyncLock(
                self.async_redis_client,
                lock_key,
                timeout=60,
                blocking=False  # 不阻塞，直接跳过
            )
            
            if not await lock.acquire():
                logger.info(f"Another process is claiming messages for {offline_consumer}")
                return 0
                
            try:
                # 获取Stream的key
                prefixed_queue = self.queue_formatter(queue)
                logger.debug(f"Stream key: {prefixed_queue}, type: {type(prefixed_queue)}")
                
                # 获取所有的consumer groups
                all_groups = await self._get_consumer_groups(prefixed_queue)
                logger.info(f"Found {len(all_groups)} groups for stream {prefixed_queue}")
                
                for group_name in all_groups:
                    # 获取该group的所有consumers
                    try:
                        consumers_info = await self.async_redis_client.xinfo_consumers(prefixed_queue, group_name)
                        logger.debug(f"Consumers in group {group_name}: {consumers_info}")
                        
                        for consumer_info in consumers_info:
                            # 二进制Redis客户端返回的字典键是字符串，值是bytes
                            consumer_name = consumer_info.get('name', b'')
                            if isinstance(consumer_name, bytes):
                                consumer_name = consumer_name.decode('utf-8')
                            
                            pending_count = consumer_info.get('pending', 0)
                            logger.info(f'{offline_consumer=} {consumer_name=}')
                            # 检查是否是离线worker的consumer
                            # 离线consumer名称格式：{consumer_id}-{queue} 或 {consumer_id}-{queue}:{task_name}
                            if consumer_name.startswith(f"{offline_consumer}"):
                                logger.info(f'{consumer_info=}')
                                if pending_count > 0:
                                    logger.info(f"Found {pending_count} pending messages for {consumer_name} in group {group_name}")
                                    
                                    # 确定新的consumer名称
                                    # 如果原consumer有task后缀，新consumer也要有相同的task后缀
                                    if ':' in consumer_name and ':' in group_name:
                                        # 提取task后缀
                                        task_suffix = consumer_name.split(':', 1)[1]
                                        # 新consumer也要有相同的task后缀
                                        if ':' in current_consumer_name:
                                            # 如果当前consumer已经有后缀，保持它
                                            new_consumer = current_consumer_name
                                        else:
                                            # 添加task后缀
                                            new_consumer = f"{current_consumer_name}:{task_suffix}"
                                    else:
                                        new_consumer = current_consumer_name
                                    
                                    logger.info(f"Claiming messages from {consumer_name} to {new_consumer}")
                                    
                                    # 转移pending消息
                                    claimed = await self._claim_messages(
                                        prefixed_queue, group_name,
                                        consumer_name, new_consumer
                                    )
                                    
                                    if claimed:
                                        total_claimed += len(claimed)
                                        logger.info(f"Successfully claimed {len(claimed)} messages")
                                        
                                        # 处理转移的消息
                                        if process_message_callback:
                                            for msg_id, msg_data in claimed:
                                                await process_message_callback(msg_id, msg_data, queue, consumer_id)
                                        elif event_queue:
                                            for msg_id, msg_data in claimed:
                                                await self._put_to_event_queue(
                                                    msg_id, msg_data, queue,
                                                    event_queue, new_consumer,
                                                    group_name, consumer_name
                                                )
                    except Exception as e:
                        logger.error(f"Error processing group {group_name}: {e}")
                
                # 标记该worker的消息已转移（即使没有消息也要标记，避免重复处理）
                await self.async_redis_client.hset(worker_key, 'messages_transferred', 'true')
                if total_claimed > 0:
                    logger.info(f"Transferred total {total_claimed} messages from worker {consumer_id}")
                else:
                    logger.info(f"No messages to transfer from worker {consumer_id}, marked as processed")
                    
            finally:
                await lock.release()
                
        except Exception as e:
            logger.error(f"Error recovering messages: {e}")
            
        return total_claimed
        
    async def _get_consumer_groups(self, stream_key: str, suffix: Optional[str] = None) -> List[str]:
        """获取Stream的consumer groups"""
        groups = []
        try:
            # 确保stream_key是字符串类型（如果是bytes会出问题）
            if isinstance(stream_key, bytes):
                stream_key = stream_key.decode('utf-8')
            
            all_groups = await self.async_redis_client.xinfo_groups(stream_key)
            logger.debug(f"Raw groups info for {stream_key}: {all_groups}")
            
            for group_info in all_groups:
                # 二进制Redis客户端返回的字典键是字符串，值是bytes
                group_name = group_info.get('name', b'')
                logger.debug(f"Processing group: {group_info}, group_name type: {type(group_name)}")
                
                # 解码group名称
                if isinstance(group_name, bytes):
                    group_name = group_name.decode('utf-8')
                
                # 过滤空的group名称
                if group_name:
                    if suffix:
                        if group_name.endswith(suffix):
                            groups.append(group_name)
                    else:
                        groups.append(group_name)
                        logger.debug(f"Added group: {group_name}")
        except Exception as e:
            logger.error(f"Error getting consumer groups for {stream_key}: {e}")
        return groups
        
    async def _claim_messages(self, stream_key: str, group_name: str, 
                             old_consumer: str, new_consumer: str) -> List[Tuple[bytes, Dict]]:
        """转移pending消息"""
        all_claimed = []
        last_id = '-'
        
        try:
            # 确保参数是bytes类型
            if isinstance(stream_key, str):
                stream_key = stream_key.encode('utf-8')
            if isinstance(group_name, str):
                group_name = group_name.encode('utf-8')
            
            logger.debug(f"_claim_messages: stream_key={stream_key}, group_name={group_name}, old_consumer={old_consumer}, new_consumer={new_consumer}")
                
            while True:
                # 获取pending消息
                pending_batch = await self.async_redis_client.xpending_range(
                    stream_key, group_name,
                    min=last_id, max='+',
                    count=100
                )
                
                logger.debug(f"Got {len(pending_batch) if pending_batch else 0} pending messages")
                
                if not pending_batch:
                    break
                
                # 过滤出属于旧consumer的消息
                message_ids = []
                for msg in pending_batch:
                    msg_consumer = msg.get('consumer') or msg.get(b'consumer')
                    if isinstance(msg_consumer, bytes):
                        msg_consumer = msg_consumer.decode('utf-8')
                    
                    if msg_consumer == old_consumer:
                        msg_id = msg.get('message_id') or msg.get(b'message_id')
                        message_ids.append(msg_id)
                
                if message_ids:
                    # 使用XCLAIM转移消息
                    logger.info(f"Claiming {len(message_ids)} messages from {old_consumer} to {new_consumer}")
                    
                    claimed = await self.async_redis_client.xclaim(
                        stream_key, group_name,
                        new_consumer,
                        min_idle_time=0,
                        message_ids=message_ids,
                        force=True
                    )
                    
                    if claimed:
                        all_claimed.extend(claimed)
                
                # 更新游标
                if pending_batch:
                    last_msg_id = pending_batch[-1].get('message_id') or pending_batch[-1].get(b'message_id')
                    if isinstance(last_msg_id, bytes):
                        last_msg_id = last_msg_id.decode('utf-8')
                    # 增加ID以获取下一批
                    parts = last_msg_id.split('-')
                    if len(parts) == 2:
                        last_id = f"{parts[0]}-{int(parts[1]) + 1}"
                    else:
                        break
                else:
                    break
                    
        except Exception as e:
            logger.error(f"Error claiming messages: {e}")
            
        return all_claimed
        
    async def _put_to_event_queue(self, msg_id, msg_data, queue, event_queue, 
                                 consumer, group_name, old_consumer):
        """将转移的消息放入event_queue"""
        try:
            # 解析消息数据
            if b'data' in msg_data:
                event_data = msgpack.unpackb(msg_data[b'data'], raw=False)
            else:
                event_data = msg_data
            
            # 构建事件
            event = {
                'event_id': msg_id.decode() if isinstance(msg_id, bytes) else msg_id,
                'event_data': event_data,
                'queue': queue,
                'consumer': consumer,
                'group_name': group_name,
                '_recovery': True,
                '_claimed_from': old_consumer
            }
            
            await event_queue.put(event)
            
        except Exception as e:
            logger.error(f"Error putting message to event queue: {e}")
            
    def stop(self):
        """停止恢复处理"""
        self._stop_recovery = True


#!/usr/bin/env python
"""简化的恢复机制测试脚本"""

import asyncio
import json
import logging
import os
import signal
import sys
import time
from typing import Dict, List, Optional
from datetime import datetime, timezone

import redis.asyncio as redis
from redis.asyncio import Redis
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from sqlalchemy.orm import sessionmaker
from sqlalchemy import text

from jettask.webui.config import PostgreSQLConfig, RedisConfig
from jettask.core.consumer_manager import ConsumerManager, ConsumerStrategy
from jettask.utils.serializer import dumps_str

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)

class RecoveryTester:
    """恢复机制测试器"""
    
    def __init__(self):
        self.pg_config = PostgreSQLConfig(
            host=os.getenv('JETTASK_PG_HOST', 'localhost'),
            port=int(os.getenv('JETTASK_PG_PORT', '5432')),
            database=os.getenv('JETTASK_PG_DB', 'jettask'),
            user=os.getenv('JETTASK_PG_USER', 'jettask'),
            password=os.getenv('JETTASK_PG_PASSWORD', '123456'),
        )
        
        self.redis_config = RedisConfig(
            host=os.getenv('REDIS_HOST', 'localhost'),
            port=int(os.getenv('REDIS_PORT', '6379')),
            db=int(os.getenv('REDIS_DB', '0')),
            password=os.getenv('REDIS_PASSWORD'),
        )
        
        self.prefix = "jettask"
        self.redis_client: Optional[Redis] = None
        self.async_engine = None
        self.AsyncSessionLocal = None
        
    async def setup(self):
        """初始化连接"""
        self.redis_client = await redis.Redis(
            host=self.redis_config.host,
            port=self.redis_config.port,
            db=self.redis_config.db,
            password=self.redis_config.password,
            decode_responses=False
        )
        
        if self.pg_config.dsn.startswith('postgresql://'):
            dsn = self.pg_config.dsn.replace('postgresql://', 'postgresql+psycopg://', 1)
        else:
            dsn = self.pg_config.dsn
            
        self.async_engine = create_async_engine(dsn, pool_size=20, echo=False)
        
        self.AsyncSessionLocal = sessionmaker(
            self.async_engine,
            class_=AsyncSession,
            expire_on_commit=False
        )
        
    async def cleanup(self):
        """清理连接"""
        if self.redis_client:
            await self.redis_client.aclose()
        if self.async_engine:
            await self.async_engine.dispose()
    
    async def test_queue_recovery(self):
        """测试队列消息的恢复机制"""
        logger.info("=" * 60)
        logger.info("测试 _consume_queues 的恢复机制")
        logger.info("=" * 60)
        
        queue_name = 'RECOVERY_TEST_QUEUE'
        stream_key = f"{self.prefix}:QUEUE:{queue_name}"
        consumer_group = f"{self.prefix}_pg_consumer1"
        
        # 1. 清理旧数据
        logger.info("\n1. 清理旧数据...")
        try:
            await self.redis_client.delete(stream_key)
        except:
            pass
        
        # 2. 发送测试消息
        logger.info("\n2. 发送测试消息到队列...")
        task_ids = []
        for i in range(5):
            task_data = {
                'name': f'recovery_test_task_{i}',
                'queue': queue_name,
                'priority': 1,
                'trigger_time': time.time(),
                'test_id': f'queue_recovery_{i}'
            }
            
            msg_id = await self.redis_client.xadd(
                stream_key,
                {'data': dumps_str(task_data)}
            )
            task_ids.append(msg_id)
            logger.info(f"  - 发送消息 {i}: {msg_id}")
        
        # 3. 创建消费者组
        logger.info("\n3. 创建消费者组...")
        try:
            await self.redis_client.xgroup_create(
                stream_key, consumer_group, id='0', mkstream=True
            )
            logger.info(f"  - 创建消费者组: {consumer_group}")
        except:
            pass
        
        # 4. 模拟consumer_1读取消息但不ACK（突然挂掉）
        logger.info("\n4. 模拟 consumer_1 读取消息但不ACK...")
        consumer_1 = "test_consumer_1"
        
        messages = await self.redis_client.xreadgroup(
            consumer_group,
            consumer_1,
            {stream_key: '>'},
            count=5,
            block=1000
        )
        
        if messages:
            logger.info(f"  - consumer_1 读取了 {len(messages[0][1])} 条消息")
            for msg_id, data in messages[0][1]:
                logger.info(f"    - 消息ID: {msg_id}")
        
        # 5. 检查pending消息
        logger.info("\n5. 检查pending消息...")
        pending_info = await self.redis_client.xpending(stream_key, consumer_group)
        logger.info(f"  - 总pending消息数: {pending_info['pending']}")
        
        if pending_info['pending'] > 0:
            detailed = await self.redis_client.xpending_range(
                stream_key, consumer_group,
                min='-', max='+', count=10
            )
            for msg in detailed:
                logger.info(f"    - 消息 {msg['message_id']}: consumer={msg['consumer']}, times_delivered={msg['times_delivered']}")
        
        # 6. 模拟consumer_2接管pending消息
        logger.info("\n6. 模拟 consumer_2 尝试接管pending消息...")
        consumer_2 = "test_consumer_2"
        
        # 使用XCLAIM接管消息
        if pending_info['pending'] > 0:
            # 获取所有pending消息的ID
            pending_msg_ids = [msg['message_id'] for msg in detailed]
            
            # XCLAIM接管消息（idle时间设为0表示立即接管）
            claimed = await self.redis_client.xclaim(
                stream_key,
                consumer_group,
                consumer_2,
                min_idle_time=0,  # 立即接管
                message_ids=pending_msg_ids
            )
            
            logger.info(f"  - consumer_2 接管了 {len(claimed)} 条消息")
            
            # ACK消息
            if claimed:
                msg_ids_to_ack = [msg[0] for msg in claimed]
                await self.redis_client.xack(stream_key, consumer_group, *msg_ids_to_ack)
                logger.info(f"  - consumer_2 ACK了 {len(msg_ids_to_ack)} 条消息")
        
        # 7. 再次检查pending消息
        logger.info("\n7. 再次检查pending消息...")
        pending_info_after = await self.redis_client.xpending(stream_key, consumer_group)
        logger.info(f"  - 总pending消息数: {pending_info_after['pending']}")
        
        # 8. 验证结果
        logger.info("\n" + "=" * 60)
        if pending_info['pending'] > 0 and pending_info_after['pending'] == 0:
            logger.info("✓ 队列消息恢复测试通过！pending消息成功被接管和处理")
        else:
            logger.warning(f"✗ 队列消息恢复测试失败。恢复前: {pending_info['pending']}，恢复后: {pending_info_after['pending']}")
        
        return pending_info['pending'] > 0 and pending_info_after['pending'] == 0
    
    async def test_task_changes_recovery(self):
        """测试TASK_CHANGES的恢复机制"""
        logger.info("\n" + "=" * 60)
        logger.info("测试 _consume_task_changes 的恢复机制")
        logger.info("=" * 60)
        
        change_stream_key = f"{self.prefix}:TASK_CHANGES"
        consumer_group = f"{self.prefix}_changes_consumer"
        
        # 1. 清理旧数据
        logger.info("\n1. 清理旧数据...")
        try:
            await self.redis_client.delete(change_stream_key)
        except:
            pass
        
        # 2. 发送任务变更事件
        logger.info("\n2. 发送任务变更事件...")
        event_ids = []
        for i in range(5):
            event_data = {
                'event_id': f'task_change_test_{i}',
                'event_type': 'task_updated',
                'timestamp': str(time.time())
            }
            
            msg_id = await self.redis_client.xadd(
                change_stream_key,
                event_data
            )
            event_ids.append(msg_id)
            logger.info(f"  - 发送事件 {i}: {msg_id}")
            
            # 同时创建对应的任务数据
            task_key = f"{self.prefix}:TASK:task_change_test_{i}"
            task_data = {
                'status': 'completed',
                'completed_at': str(time.time()),
                'result': json.dumps({'success': True})
            }
            await self.redis_client.hset(task_key, mapping={
                k: dumps_str(v) if not isinstance(v, (str, bytes)) else v
                for k, v in task_data.items()
            })
        
        # 3. 创建消费者组
        logger.info("\n3. 创建消费者组...")
        try:
            await self.redis_client.xgroup_create(
                change_stream_key, consumer_group, id='0', mkstream=True
            )
            logger.info(f"  - 创建消费者组: {consumer_group}")
        except:
            pass
        
        # 4. 模拟consumer_1读取消息但不ACK
        logger.info("\n4. 模拟 consumer_1 读取消息但不ACK...")
        consumer_1 = "changes_consumer_1"
        
        messages = await self.redis_client.xreadgroup(
            consumer_group,
            consumer_1,
            {change_stream_key: '>'},
            count=5,
            block=1000
        )
        
        if messages:
            logger.info(f"  - consumer_1 读取了 {len(messages[0][1])} 条事件")
            for msg_id, data in messages[0][1]:
                logger.info(f"    - 事件ID: {msg_id}")
        
        # 5. 检查pending消息
        logger.info("\n5. 检查pending事件...")
        pending_info = await self.redis_client.xpending(change_stream_key, consumer_group)
        logger.info(f"  - 总pending事件数: {pending_info['pending']}")
        
        if pending_info['pending'] > 0:
            detailed = await self.redis_client.xpending_range(
                change_stream_key, consumer_group,
                min='-', max='+', count=10
            )
            for msg in detailed:
                logger.info(f"    - 事件 {msg['message_id']}: consumer={msg['consumer']}, times_delivered={msg['times_delivered']}")
        
        # 6. 模拟consumer_2接管pending消息
        logger.info("\n6. 模拟 consumer_2 接管pending事件...")
        consumer_2 = "changes_consumer_2"
        
        if pending_info['pending'] > 0:
            pending_msg_ids = [msg['message_id'] for msg in detailed]
            
            claimed = await self.redis_client.xclaim(
                change_stream_key,
                consumer_group,
                consumer_2,
                min_idle_time=0,
                message_ids=pending_msg_ids
            )
            
            logger.info(f"  - consumer_2 接管了 {len(claimed)} 条事件")
            
            # ACK消息
            if claimed:
                msg_ids_to_ack = [msg[0] for msg in claimed]
                await self.redis_client.xack(change_stream_key, consumer_group, *msg_ids_to_ack)
                logger.info(f"  - consumer_2 ACK了 {len(msg_ids_to_ack)} 条事件")
        
        # 7. 再次检查pending消息
        logger.info("\n7. 再次检查pending事件...")
        pending_info_after = await self.redis_client.xpending(change_stream_key, consumer_group)
        logger.info(f"  - 总pending事件数: {pending_info_after['pending']}")
        
        # 8. 验证结果
        logger.info("\n" + "=" * 60)
        if pending_info['pending'] > 0 and pending_info_after['pending'] == 0:
            logger.info("✓ TASK_CHANGES恢复测试通过！pending事件成功被接管和处理")
        else:
            logger.warning(f"✗ TASK_CHANGES恢复测试失败。恢复前: {pending_info['pending']}，恢复后: {pending_info_after['pending']}")
        
        return pending_info['pending'] > 0 and pending_info_after['pending'] == 0
    
    async def test_offline_worker_recovery(self):
        """测试OfflineWorkerRecovery的功能"""
        logger.info("\n" + "=" * 60)
        logger.info("测试 OfflineWorkerRecovery 的功能")
        logger.info("=" * 60)
        
        # 导入OfflineWorkerRecovery
        from jettask.core.offline_worker_recovery import OfflineWorkerRecovery
        
        # 创建同步Redis客户端（ConsumerManager需要）
        import redis as sync_redis
        sync_redis_client = sync_redis.StrictRedis(
            host=self.redis_config.host,
            port=self.redis_config.port,
            db=self.redis_config.db,
            password=self.redis_config.password,
            decode_responses=True
        )
        
        # 创建ConsumerManager
        consumer_manager = ConsumerManager(
            redis_client=sync_redis_client,
            strategy=ConsumerStrategy.HEARTBEAT,
            config={
                'redis_prefix': self.prefix,
                'queues': ['OFFLINE_TEST_QUEUE'],
                'worker_prefix': 'TEST_WORKER'
            }
        )
        
        # 创建OfflineWorkerRecovery
        recovery = OfflineWorkerRecovery(
            async_redis_client=self.redis_client,
            redis_prefix=self.prefix,
            worker_prefix='TEST_WORKER',
            consumer_manager=consumer_manager
        )
        
        # 准备测试数据
        queue_name = 'OFFLINE_TEST_QUEUE'
        stream_key = f"{self.prefix}:QUEUE:{queue_name}"
        consumer_group = f"{self.prefix}_pg_consumer1"
        
        # 清理旧数据
        logger.info("\n1. 清理旧数据...")
        try:
            await self.redis_client.delete(stream_key)
        except:
            pass
        
        # 发送测试消息
        logger.info("\n2. 发送测试消息...")
        for i in range(3):
            task_data = {
                'name': f'offline_test_{i}',
                'queue': queue_name,
                'test_id': f'offline_{i}'
            }
            await self.redis_client.xadd(
                stream_key,
                {'data': dumps_str(task_data)}
            )
        
        # 创建消费者组
        try:
            await self.redis_client.xgroup_create(
                stream_key, consumer_group, id='0', mkstream=True
            )
        except:
            pass
        
        # 模拟离线worker读取消息
        logger.info("\n3. 模拟离线worker读取消息...")
        offline_worker = "offline_worker_1"
        messages = await self.redis_client.xreadgroup(
            consumer_group,
            offline_worker,
            {stream_key: '>'},
            count=3,
            block=1000
        )
        
        if messages:
            logger.info(f"  - 离线worker读取了 {len(messages[0][1])} 条消息")
        
        # 模拟worker离线（删除其注册信息）
        worker_key = f"{self.prefix}:TEST_WORKER:{offline_worker}"
        await self.redis_client.delete(worker_key)
        logger.info(f"  - 模拟worker离线: 删除 {worker_key}")
        
        # 定义处理回调
        processed_messages = []
        async def process_callback(msg_id, msg_data, queue, consumer_id):
            processed_messages.append({
                'msg_id': msg_id,
                'queue': queue,
                'consumer_id': consumer_id
            })
            logger.info(f"  - 处理恢复的消息: {msg_id} from {consumer_id}")
        
        # 执行恢复
        logger.info("\n4. 执行离线worker恢复...")
        current_consumer = "active_worker_1"
        recovered_count = await recovery.recover_offline_workers(
            queue=queue_name,
            current_consumer_name=current_consumer,
            process_message_callback=process_callback
        )
        
        logger.info(f"  - 恢复了 {recovered_count} 条消息")
        
        # 检查pending消息
        logger.info("\n5. 检查恢复后的pending消息...")
        pending_info = await self.redis_client.xpending(stream_key, consumer_group)
        logger.info(f"  - 剩余pending消息数: {pending_info['pending']}")
        
        # 验证结果
        logger.info("\n" + "=" * 60)
        if recovered_count > 0 and len(processed_messages) == recovered_count:
            logger.info("✓ OfflineWorkerRecovery测试通过！成功恢复离线worker的消息")
        else:
            logger.warning(f"✗ OfflineWorkerRecovery测试失败。恢复数量: {recovered_count}，处理数量: {len(processed_messages)}")
        
        # 清理
        consumer_manager.cleanup()
        sync_redis_client.close()
        
        return recovered_count > 0


async def main():
    """主测试函数"""
    from dotenv import load_dotenv
    load_dotenv()
    
    tester = RecoveryTester()
    await tester.setup()
    
    try:
        # 测试1：队列消息恢复
        queue_test = await tester.test_queue_recovery()
        await asyncio.sleep(1)
        
        # 测试2：TASK_CHANGES恢复
        changes_test = await tester.test_task_changes_recovery()
        await asyncio.sleep(1)
        
        # 测试3：OfflineWorkerRecovery
        offline_test = await tester.test_offline_worker_recovery()
        
        # 总结
        logger.info("\n" + "=" * 60)
        logger.info("测试总结")
        logger.info("=" * 60)
        logger.info(f"队列消息恢复测试: {'✓ 通过' if queue_test else '✗ 失败'}")
        logger.info(f"TASK_CHANGES恢复测试: {'✓ 通过' if changes_test else '✗ 失败'}")
        logger.info(f"OfflineWorkerRecovery测试: {'✓ 通过' if offline_test else '✗ 失败'}")
        
        if queue_test and changes_test and offline_test:
            logger.info("\n🎉 所有测试通过！恢复机制工作正常")
        else:
            logger.warning("\n⚠️ 部分测试失败，需要检查恢复机制")
        
    except Exception as e:
        logger.error(f"测试出错: {e}", exc_info=True)
    finally:
        await tester.cleanup()


if __name__ == '__main__':
    asyncio.run(main())


"""SQLAlchemy models for JetTask WebUI database."""
from datetime import datetime
from typing import Optional, Dict, Any, List
from sqlalchemy import (
    Column, String, Integer, Float, DateTime, Text, JSON, 
    ARRAY, UniqueConstraint, Index, func
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import JSONB

Base = declarative_base()


class Task(Base):
    """任务表模型"""
    __tablename__ = 'tasks'
    
    id = Column(String(255), primary_key=True)  # Redis Stream的事件ID
    queue_name = Column(String(255), nullable=False)
    task_name = Column(String(255), nullable=False)
    task_data = Column(JSONB)  # 任务的原始数据
    priority = Column(Integer, default=0)
    retry_count = Column(Integer, default=0)
    max_retry = Column(Integer, default=3)
    status = Column(String(50), default='pending')  # pending, running, success, failed, timeout
    result = Column(JSONB)  # 执行结果
    error_message = Column(Text)
    created_at = Column(DateTime(timezone=True), default=func.current_timestamp())
    started_at = Column(DateTime(timezone=True))
    completed_at = Column(DateTime(timezone=True))
    worker_id = Column(String(255))
    execution_time = Column(Float)  # 任务执行时间（秒）
    duration = Column(Float)  # 任务总持续时间（秒）
    task_metadata = Column('metadata', JSONB)  # 额外的元数据，在数据库中仍叫metadata
    
    __table_args__ = (
        Index('idx_tasks_queue_name', 'queue_name'),
        Index('idx_tasks_status', 'status'),
        # 组合索引：优化按队列和状态查询
        Index('idx_tasks_queue_status', 'queue_name', 'status'),
        # 时间索引：优化时间范围查询
        Index('idx_tasks_created_at', 'created_at'),
        # Worker索引：优化查询特定worker的任务
        Index('idx_tasks_worker_id', 'worker_id', 
              postgresql_where=(worker_id.isnot(None))),
    )


# QueueStats 和 Worker 表已废弃，不再使用


"""
App importer module inspired by Uvicorn's import logic.
支持多种方式导入 Jettask 应用实例。
"""

import os
import sys
import importlib
import importlib.util
from pathlib import Path
from typing import Optional, Any
import inspect


class AppImporter:
    """应用导入器，支持多种导入方式"""
    
    # 默认查找的应用实例名称
    DEFAULT_APP_NAMES = ['app', 'application', 'jettask_app']
    
    # 默认查找的文件
    DEFAULT_FILES = ['app.py', 'main.py', 'server.py', 'worker.py']
    
    @classmethod
    def import_from_string(cls, import_str: str) -> Any:
        """
        从字符串导入应用，支持多种格式：
        - "module:app" - 从 module 导入 app 实例
        - "module" - 从 module 自动查找应用实例
        - "path/to/file.py:app" - 从文件路径导入
        - "path/to/file.py" - 从文件自动查找
        """
        # 分离模块路径和应用名称
        if ':' in import_str:
            module_str, app_name = import_str.rsplit(':', 1)
        else:
            module_str = import_str
            app_name = None
        
        # 判断是文件路径还是模块名
        if '/' in module_str or module_str.endswith('.py'):
            module = cls._import_from_file(module_str)
        else:
            module = cls._import_from_module(module_str)
        
        # 获取应用实例
        if app_name:
            # 支持嵌套属性访问，如 "module:app.factory()"
            if '.' in app_name or '(' in app_name:
                app = cls._evaluate_app_expression(module, app_name)
            else:
                app = getattr(module, app_name)
        else:
            # 自动查找应用实例
            app = cls._find_app_in_module(module)
            if not app:
                raise ImportError(
                    f"Cannot find Jettask app in {module_str}. "
                    f"Tried names: {', '.join(cls.DEFAULT_APP_NAMES)}"
                )
        
        return app
    
    @classmethod
    def _import_from_file(cls, file_path: str):
        """从文件路径导入模块"""
        path = Path(file_path)
        
        # 处理相对路径
        if not path.is_absolute():
            path = Path.cwd() / path
        
        # 去掉 .py 后缀（如果有）
        if path.suffix == '.py':
            path = path.with_suffix('')
        
        # 检查文件是否存在
        py_file = path.with_suffix('.py')
        if not py_file.exists():
            raise ImportError(f"File not found: {py_file}")
        
        # 添加到 sys.path
        sys.path.insert(0, str(path.parent))
        
        try:
            # 导入模块
            module_name = path.name
            spec = importlib.util.spec_from_file_location(
                module_name, 
                str(py_file)
            )
            if spec and spec.loader:
                module = importlib.util.module_from_spec(spec)
                sys.modules[module_name] = module
                try:
                    spec.loader.exec_module(module)
                    return module
                except Exception as e:
                    # 执行模块时出错，提供详细信息
                    import traceback
                    tb_lines = traceback.format_exc().splitlines()
                    
                    # 找出用户代码中的错误位置
                    user_error_lines = [line for line in tb_lines 
                                       if str(py_file) in line or module_name in line]
                    
                    raise ImportError(
                        f"Failed to load module from {py_file}.\n"
                        f"Error in user code: {str(e)}\n"
                        f"Check the file for syntax errors or missing dependencies."
                    ) from e
            else:
                raise ImportError(f"Cannot create module spec for {py_file}")
        finally:
            # 清理 sys.path
            if str(path.parent) in sys.path:
                sys.path.remove(str(path.parent))
    
    @classmethod
    def _import_from_module(cls, module_str: str):
        """从模块名导入"""
        try:
            return importlib.import_module(module_str)
        except ImportError as e:
            # 保存原始错误信息
            original_error = str(e)
            
            # 如果导入失败，尝试添加当前目录到 sys.path
            if '.' not in sys.path:
                sys.path.insert(0, '.')
                try:
                    return importlib.import_module(module_str)
                except ImportError as retry_error:
                    # 提供更详细的错误信息
                    import traceback
                    tb_str = traceback.format_exc()
                    
                    # 判断是模块不存在还是模块内部有错误
                    if f"No module named '{module_str}'" in str(retry_error):
                        raise ImportError(
                            f"Module '{module_str}' not found. "
                            f"Make sure the module exists and is in the Python path."
                        ) from retry_error
                    else:
                        # 模块存在但内部有错误，保留完整的错误链
                        raise ImportError(
                            f"Failed to import module '{module_str}'. "
                            f"The module was found but contains errors:\n{original_error}"
                        ) from retry_error
            raise
    
    @classmethod
    def _find_app_in_module(cls, module) -> Optional[Any]:
        """在模块中查找 Jettask 应用实例"""
        from ..core.app import Jettask
        
        # 按优先级查找
        for name in cls.DEFAULT_APP_NAMES:
            if hasattr(module, name):
                obj = getattr(module, name)
                if isinstance(obj, Jettask):
                    return obj
        
        # 如果没找到，遍历所有属性查找第一个 Jettask 实例
        for name in dir(module):
            if not name.startswith('_'):
                obj = getattr(module, name, None)
                if isinstance(obj, Jettask):
                    return obj
        
        return None
    
    @classmethod
    def _evaluate_app_expression(cls, module, expression: str):
        """
        评估应用表达式，支持：
        - app.factory() - 调用工厂函数
        - app.create_app() - 调用方法
        - app.instance - 访问属性
        """
        # 安全评估表达式
        # 创建受限的命名空间
        namespace = {'__builtins__': {}}
        namespace.update(vars(module))
        
        try:
            # 使用 eval 但限制在模块的命名空间内
            return eval(expression, namespace)
        except Exception as e:
            raise ImportError(f"Cannot evaluate expression '{expression}': {e}")
    
    @classmethod
    def auto_discover(cls) -> Optional[Any]:
        """
        自动发现应用，按以下顺序查找：
        1. 环境变量 JETTASK_APP
        2. 当前目录的默认文件（app.py, main.py 等）
        3. 当前目录的任何 .py 文件中的 app 实例
        """
        from ..core.app import Jettask
        
        # 1. 检查环境变量
        env_app = os.getenv('JETTASK_APP')
        if env_app:
            try:
                return cls.import_from_string(env_app)
            except Exception:
                pass
        
        # 2. 检查默认文件
        for filename in cls.DEFAULT_FILES:
            file_path = Path.cwd() / filename
            if file_path.exists():
                try:
                    module = cls._import_from_file(str(file_path))
                    app = cls._find_app_in_module(module)
                    if app:
                        return app
                except Exception:
                    continue
        
        # 3. 扫描当前目录的所有 Python 文件
        for py_file in Path.cwd().glob('*.py'):
            if py_file.name.startswith('_'):
                continue
            try:
                module = cls._import_from_file(str(py_file))
                app = cls._find_app_in_module(module)
                if app:
                    return app
            except Exception:
                continue
        
        return None
    
    @classmethod
    def get_app_info(cls, app) -> dict:
        """获取应用信息"""
        from ..core.app import Jettask
        
        if not isinstance(app, Jettask):
            return {'error': 'Not a Jettask instance'}
        
        info = {
            'type': 'Jettask',
            'redis_url': getattr(app, 'redis_url', 'Not configured'),
            'redis_prefix': getattr(app, 'redis_prefix', 'jettask'),
            'tasks': len(getattr(app, '_tasks', {})),
            'queues': [],
        }
        
        # 安全地获取队列信息
        if hasattr(app, 'ep') and hasattr(app.ep, 'queues'):
            queues = app.ep.queues
            if queues is not None:
                info['queues'] = list(queues)
        
        # 获取任务列表
        if hasattr(app, '_tasks'):
            info['task_names'] = list(app._tasks.keys())
        
        return info


def import_app(import_str: Optional[str] = None) -> Any:
    """
    导入 Jettask 应用的便捷函数。
    
    Args:
        import_str: 导入字符串，如 "module:app" 或 "path/to/file.py:app"
                   如果为 None，会尝试自动发现
    
    Returns:
        Jettask 应用实例
    
    Examples:
        >>> app = import_app("myapp:app")
        >>> app = import_app("src/main.py:create_app()")
        >>> app = import_app()  # 自动发现
    """
    if import_str:
        return AppImporter.import_from_string(import_str)
    else:
        app = AppImporter.auto_discover()
        if not app:
            raise ImportError(
                "Cannot auto-discover Jettask app. "
                "Please specify app location (e.g., 'module:app') "
                "or set JETTASK_APP environment variable."
            )
        return app


"""
独立的数据访问模块，不依赖 integrated_gradio_app.py
"""
import asyncio
import json
import logging
from datetime import datetime, timedelta, timezone
from typing import Dict, List, Optional, Tuple
import redis.asyncio as redis
from sqlalchemy import text
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from sqlalchemy.orm import sessionmaker

# 设置日志
logger = logging.getLogger(__name__)


class RedisConfig:
    """Redis配置"""
    def __init__(self, host='localhost', port=6379, db=0, password=None):
        self.host = host
        self.port = port
        self.db = db
        self.password = password
    
    @classmethod
    def from_env(cls):
        import os
        return cls(
            host=os.getenv('REDIS_HOST', 'localhost'),
            port=int(os.getenv('REDIS_PORT', 6379)),
            db=int(os.getenv('REDIS_DB', 0)),
            password=os.getenv('REDIS_PASSWORD')
        )


class PostgreSQLConfig:
    """PostgreSQL配置"""
    def __init__(self, host='localhost', port=5432, user='postgres', password='', database='jettask'):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
    
    @property
    def dsn(self):
        return f"postgresql://{self.user}:{self.password}@{self.host}:{self.port}/{self.database}"
    
    @classmethod
    def from_env(cls):
        import os
        return cls(
            host=os.getenv('POSTGRES_HOST', 'localhost'),
            port=int(os.getenv('POSTGRES_PORT', 5432)),
            user=os.getenv('POSTGRES_USER', 'jettask'),
            password=os.getenv('POSTGRES_PASSWORD', '123456'),
            database=os.getenv('POSTGRES_DB', 'jettask')
        )


class JetTaskDataAccess:
    """JetTask数据访问类"""
    
    def __init__(self):
        self.redis_config = RedisConfig.from_env()
        self.pg_config = PostgreSQLConfig.from_env()
        self.redis_prefix = "jettask"
        self.async_engine = None
        self.AsyncSessionLocal = None
        self._redis_pool = None
        
    async def initialize(self):
        """初始化数据库连接"""
        try:
            # 初始化PostgreSQL引擎
            dsn = self.pg_config.dsn
            if dsn.startswith('postgresql://'):
                dsn = dsn.replace('postgresql://', 'postgresql+psycopg://', 1)
            
            self.async_engine = create_async_engine(
                dsn,
                pool_size=10,
                max_overflow=5,
                pool_pre_ping=True,
                echo=False
            )
            
            self.AsyncSessionLocal = sessionmaker(
                bind=self.async_engine,
                class_=AsyncSession,
                expire_on_commit=False
            )
            
            # 初始化Redis连接池
            self._redis_pool = redis.ConnectionPool(
                host=self.redis_config.host,
                port=self.redis_config.port,
                db=self.redis_config.db,
                password=self.redis_config.password,
                encoding='utf-8',
                decode_responses=True
            )
            
            logger.info("数据库连接初始化成功")
            
        except Exception as e:
            logger.error(f"数据库连接初始化失败: {e}")
            raise
    
    async def close(self):
        """关闭数据库连接"""
        if self.async_engine:
            await self.async_engine.dispose()
        if self._redis_pool:
            await self._redis_pool.disconnect()
    
    async def get_redis_client(self):
        """获取Redis客户端"""
        return redis.Redis(connection_pool=self._redis_pool)
    
    async def fetch_queues_data(self) -> List[Dict]:
        """获取队列数据（基于Redis Stream）"""
        try:
            redis_client = await self.get_redis_client()
            
            # 获取所有Stream类型的队列 - JetTask使用 jettask:QUEUE:队列名 格式
            all_keys = await redis_client.keys(f"{self.redis_prefix}:QUEUE:*")
            queues_data = []
            queue_names = set()
            
            for key in all_keys:
                # 检查是否是Stream类型
                key_type = await redis_client.type(key)
                if key_type == 'stream':
                    # 解析队列名称 - 格式: jettask:QUEUE:队列名
                    parts = key.split(':')
                    if len(parts) >= 3 and parts[0] == self.redis_prefix and parts[1] == 'QUEUE':
                        queue_name = ':'.join(parts[2:])  # 支持带冒号的队列名
                        queue_names.add(queue_name)
            
            # 获取每个队列的详细信息
            for queue_name in queue_names:
                stream_key = f"{self.redis_prefix}:QUEUE:{queue_name}"
                
                try:
                    # 获取Stream信息
                    stream_info = await redis_client.xinfo_stream(stream_key)
                    
                    # 获取消费者组信息
                    groups_info = await redis_client.xinfo_groups(stream_key)
                    
                    pending_count = 0
                    processing_count = 0
                    
                    # 统计各消费者组的待处理消息
                    for group in groups_info:
                        pending_count += group.get('pending', 0)
                        
                        # 获取消费者信息
                        try:
                            consumers = await redis_client.xinfo_consumers(stream_key, group['name'])
                            for consumer in consumers:
                                processing_count += consumer.get('pending', 0)
                        except:
                            pass
                    
                    # Stream的长度即为总消息数
                    total_messages = stream_info.get('length', 0)
                    
                    # 完成的消息数 = 总消息数 - 待处理 - 处理中
                    completed_count = max(0, total_messages - pending_count - processing_count)
                    
                    queues_data.append({
                        '队列名称': queue_name,
                        '待处理': pending_count,
                        '处理中': processing_count,
                        '已完成': completed_count,
                        '失败': 0,  # Stream中没有直接的失败计数
                        '总计': total_messages
                    })
                    
                except Exception as e:
                    logger.warning(f"获取队列 {queue_name} 信息失败: {e}")
                    # 如果获取详细信息失败，至少返回队列名称
                    queues_data.append({
                        '队列名称': queue_name,
                        '待处理': 0,
                        '处理中': 0,
                        '已完成': 0,
                        '失败': 0,
                        '总计': 0
                    })
            
            await redis_client.close()
            return sorted(queues_data, key=lambda x: x['队列名称'])
            
        except Exception as e:
            logger.error(f"获取队列数据失败: {e}")
            return []
    
    async def fetch_queue_timeline_data(self, 
                                      queues: List[str], 
                                      start_time: datetime, 
                                      end_time: datetime) -> List[Dict]:
        """获取队列时间线数据 - 返回真实的任务触发时间"""
        try:
            await asyncio.sleep(1)
            if not self.AsyncSessionLocal:
                await self.initialize()
            
            async with self.AsyncSessionLocal() as session:
                # 构建SQL查询
                queue_names_str = "', '".join(queues)
                
                # 计算时间跨度
                duration = (end_time - start_time).total_seconds()
                
                # 根据时间跨度决定聚合粒度
                # 短时间范围：返回每个任务的真实时间
                # 长时间范围：按适当的时间窗口聚合
                
                if duration <= 3600:  # 1小时以内，返回每个任务的真实时间
                    print(f'一小时内')
                    # 查询每个任务的真实创建时间
                    query = text(f"""
                        SELECT 
                            created_at as time_point,
                            queue_name,
                            COUNT(*) OVER (
                                PARTITION BY queue_name, created_at
                            ) as task_count
                        FROM tasks 
                        WHERE queue_name IN ('{queue_names_str}')
                            AND created_at >= :start_time 
                            AND created_at <= :end_time
                        ORDER BY created_at, queue_name
                    """)
                    
                elif duration <= 86400:  # 1天以内，按分钟聚合
                    query = text(f"""
                        SELECT 
                            date_trunc('minute', created_at) as time_point,
                            queue_name,
                            COUNT(*) as task_count
                        FROM tasks 
                        WHERE queue_name IN ('{queue_names_str}')
                            AND created_at >= :start_time 
                            AND created_at <= :end_time
                        GROUP BY time_point, queue_name
                        ORDER BY time_point, queue_name
                    """)
                    
                elif duration <= 604800:  # 7天以内，按小时聚合
                    query = text(f"""
                        SELECT 
                            date_trunc('hour', created_at) as time_point,
                            queue_name,
                            COUNT(*) as task_count
                        FROM tasks 
                        WHERE queue_name IN ('{queue_names_str}')
                            AND created_at >= :start_time 
                            AND created_at <= :end_time
                        GROUP BY time_point, queue_name
                        ORDER BY time_point, queue_name
                    """)
                    
                else:  # 超过7天，按天聚合
                    query = text(f"""
                        SELECT 
                            date_trunc('day', created_at) as time_point,
                            queue_name,
                            COUNT(*) as task_count
                        FROM tasks 
                        WHERE queue_name IN ('{queue_names_str}')
                            AND created_at >= :start_time 
                            AND created_at <= :end_time
                        GROUP BY time_point, queue_name
                        ORDER BY time_point, queue_name
                    """)
                
                result = await session.execute(query, {
                    'start_time': start_time,
                    'end_time': end_time
                })
                
                # 处理查询结果，返回真实的时间点
                timeline_data = []
                seen_points = set()  # 用于去重（对于窗口函数的结果）
                
                for row in result:
                    # 创建唯一键避免重复
                    unique_key = f"{row.time_point.isoformat()}_{row.queue_name}"
                    if unique_key not in seen_points:
                        seen_points.add(unique_key)
                        timeline_data.append({
                            'time': row.time_point.isoformat(),
                            'queue': row.queue_name,
                            'value': row.task_count
                        })
                
                logger.info(f"从DB查询到 {len(timeline_data)} 个真实数据点")
                
                return timeline_data
                
        except Exception as e:
            logger.error(f"获取队列时间线数据失败: {e}")
            return []
    
    
    async def fetch_global_stats(self) -> Dict:
        """获取全局统计信息"""
        try:
            redis_client = await self.get_redis_client()
            
            # 获取所有队列的统计信息
            queues_data = await self.fetch_queues_data()
            
            total_pending = sum(q['待处理'] for q in queues_data)
            total_processing = sum(q['处理中'] for q in queues_data)
            total_completed = sum(q['已完成'] for q in queues_data)
            total_failed = sum(q['失败'] for q in queues_data)
            
            # 获取活跃worker数量
            worker_keys = await redis_client.keys(f"{self.redis_prefix}:worker:*")
            active_workers = len(worker_keys)
            
            await redis_client.close()
            
            return {
                'total_queues': len(queues_data),
                'total_pending': total_pending,
                'total_processing': total_processing,
                'total_completed': total_completed,
                'total_failed': total_failed,
                'active_workers': active_workers,
                'timestamp': datetime.now(timezone.utc).isoformat()
            }
            
        except Exception as e:
            logger.error(f"获取全局统计信息失败: {e}")
            return {}


from .core.app import Jettask
from .core.task import Task, Request, ExecuteResponse
from .core.context import TaskContext
from .router import TaskRouter
from .exceptions import JetTaskException, TaskTimeoutError, TaskExecutionError, TaskNotFoundError, RetryableError
from .utils.error_handler import clean_task_errors, handle_task_error
from .utils.task_logger import (
    configure_task_logging,
    get_task_logger,
    TaskContextManager,
    set_task_context,
    clear_task_context,
    LogContext
)

# 自动安装简化的异常显示（用户可通过环境变量 JETTASK_FULL_TRACEBACK=1 禁用）
from .utils import exception_hook  # 这会自动安装 excepthook

__version__ = "0.1.0"
__all__ = [
    "Jettask", 
    "Task",
    "TaskContext",
    "TaskRouter", 
    "Request", 
    "ExecuteResponse",
    "JetTaskException",
    "TaskTimeoutError",
    "TaskExecutionError",
    "TaskNotFoundError",
    "RetryableError",
    "clean_task_errors",
    "handle_task_error",
    "configure_task_logging",
    "get_task_logger",
    "TaskContextManager",
    "set_task_context",
    "clear_task_context",
    "LogContext"
]


#!/usr/bin/env python
"""测试版本的 PostgreSQL Consumer - 用于测试 pending 消息恢复"""

import asyncio
import json
import logging
import os
import time
import signal
from typing import Dict, List, Optional, Any
from datetime import datetime, timezone
from collections import defaultdict

import redis.asyncio as redis
from redis.asyncio import Redis
from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from sqlalchemy.orm import sessionmaker
from sqlalchemy import text

from jettask.webui.config import PostgreSQLConfig, RedisConfig
from jettask.core.consumer_manager import ConsumerManager, ConsumerStrategy
from jettask.core.offline_worker_recovery import OfflineWorkerRecovery

logger = logging.getLogger(__name__)


class TestPostgreSQLConsumer:
    """测试版本的PostgreSQL消费者，用于测试pending消息恢复"""
    
    def __init__(self, pg_config: PostgreSQLConfig, redis_config: RedisConfig, 
                 prefix: str = "jettask", node_id: str = None, 
                 consumer_strategy: ConsumerStrategy = ConsumerStrategy.HEARTBEAT,
                 test_mode: str = None,
                 processing_delay: float = 0,
                 crash_after_messages: int = 0):
        """
        Args:
            test_mode: 测试模式 
                - "slow_process": 模拟缓慢处理
                - "crash_after_n": 处理N条消息后崩溃
                - "normal": 正常处理
            processing_delay: 每条消息的处理延迟（秒）
            crash_after_messages: 处理多少条消息后崩溃（仅在crash_after_n模式下有效）
        """
        self.pg_config = pg_config
        self.redis_config = redis_config
        self.prefix = prefix
        self.redis_client: Optional[Redis] = None
        self.async_engine = None
        self.AsyncSessionLocal = None
        self.consumer_group = f"{prefix}_pg_consumer1"
        
        # 测试配置
        self.test_mode = test_mode or "normal"
        self.processing_delay = processing_delay
        self.crash_after_messages = crash_after_messages
        self.processed_count = 0
        self.queue_processed_count = defaultdict(int)
        self.changes_processed_count = 0
        
        # 节点标识
        import socket
        hostname = socket.gethostname()
        self.node_id = node_id or f"{hostname}_{os.getpid()}"
        
        # 使用 ConsumerManager 来管理 consumer_id
        self.consumer_strategy = consumer_strategy
        self.consumer_manager = None
        self.consumer_id = None
        
        self._running = False
        self._tasks = []
        self._known_queues = set()
        self._consecutive_errors = defaultdict(int)
        
        # 内存中维护已处理的任务ID集合
        self._processed_task_ids = set()
        self._processed_ids_lock = asyncio.Lock()
        self._processed_ids_max_size = 100000
        
        # 待重试的任务更新
        self._pending_updates = {}
        self._pending_updates_lock = asyncio.Lock()
        self._max_pending_updates = 10000
        self._retry_interval = 5
        
        # 动态批次大小
        self.batch_size = 100  # 减小批次便于测试
        self.min_batch_size = 10
        self.max_batch_size = 200
        
    async def start(self):
        """启动消费者"""
        logger.info(f"Starting TEST PostgreSQL consumer on node: {self.node_id}")
        logger.info(f"Test mode: {self.test_mode}, delay: {self.processing_delay}s, crash after: {self.crash_after_messages}")
        
        # 连接Redis
        self.redis_client = await redis.Redis(
            host=self.redis_config.host,
            port=self.redis_config.port,
            db=self.redis_config.db,
            password=self.redis_config.password,
            decode_responses=False
        )
        
        # 初始化 ConsumerManager
        import redis as sync_redis
        sync_redis_client = sync_redis.StrictRedis(
            host=self.redis_config.host,
            port=self.redis_config.port,
            db=self.redis_config.db,
            password=self.redis_config.password,
            decode_responses=True
        )
        
        # 配置 ConsumerManager
        initial_queues = ['TASK_CHANGES']
        consumer_config = {
            'redis_prefix': self.prefix,
            'queues': initial_queues,
            'worker_prefix': 'PG_CONSUMER',
        }
        
        self.consumer_manager = ConsumerManager(
            redis_client=sync_redis_client,
            strategy=self.consumer_strategy,
            config=consumer_config
        )
        
        # 获取稳定的 consumer_id
        self.consumer_id = self.consumer_manager.get_consumer_name('TASK_CHANGES')
        logger.info(f"Using consumer_id: {self.consumer_id} with strategy: {self.consumer_strategy.value}")
        
        # 创建SQLAlchemy异步引擎
        if self.pg_config.dsn.startswith('postgresql://'):
            dsn = self.pg_config.dsn.replace('postgresql://', 'postgresql+psycopg://', 1)
        else:
            dsn = self.pg_config.dsn
            
        self.async_engine = create_async_engine(
            dsn,
            pool_size=10,
            max_overflow=5,
            pool_pre_ping=True,
            pool_recycle=300,
            echo=False
        )
        
        # 创建异步会话工厂
        self.AsyncSessionLocal = sessionmaker(
            self.async_engine,
            class_=AsyncSession,
            expire_on_commit=False
        )
        
        # 初始化数据库架构
        await self._init_database()
        
        self._running = True
        
        # 先进行一次队列发现
        await self._initial_queue_discovery()
        
        # 创建离线worker恢复器
        self.offline_recovery = OfflineWorkerRecovery(
            async_redis_client=self.redis_client,
            redis_prefix=self.prefix,
            worker_prefix='PG_CONSUMER',
            consumer_manager=self.consumer_manager
        )
        
        # 启动消费任务（简化版：只保留必要的任务）
        self._tasks = [
            asyncio.create_task(self._consume_queues()),           # 消费新任务
            asyncio.create_task(self._consume_task_changes()),     # 消费任务变更事件
            asyncio.create_task(self._start_offline_recovery())    # 离线worker恢复服务
        ]
        
        logger.info("TEST PostgreSQL consumer started successfully")
        
    async def stop(self):
        """停止消费者"""
        logger.info("Stopping TEST PostgreSQL consumer...")
        self._running = False
        
        # 停止离线恢复服务
        if hasattr(self, 'offline_recovery'):
            self.offline_recovery.stop()
        
        # 取消所有任务
        for task in self._tasks:
            task.cancel()
        
        # 等待任务完成
        await asyncio.gather(*self._tasks, return_exceptions=True)
        
        # 清理 ConsumerManager
        if self.consumer_manager:
            try:
                self.consumer_manager.cleanup()
                logger.info(f"Cleaned up ConsumerManager for consumer: {self.consumer_id}")
            except Exception as e:
                logger.error(f"Error cleaning up ConsumerManager: {e}")
        
        # 关闭连接
        if self.redis_client:
            await self.redis_client.close()
        
        if self.async_engine:
            await self.async_engine.dispose()
            
        logger.info("TEST PostgreSQL consumer stopped")
        
    async def _init_database(self):
        """初始化数据库架构"""
        import os
        current_dir = os.path.dirname(os.path.abspath(__file__))
        schema_path = os.path.join(current_dir, "schema.sql")
        try:
            with open(schema_path, 'r') as f:
                schema_sql = f.read()
            
            async with self.AsyncSessionLocal() as session:
                await session.execute(text(schema_sql))
                await session.commit()
                logger.info("Database schema initialized")
        except FileNotFoundError:
            logger.warning(f"Schema file not found at {schema_path}, skipping initialization")
        except Exception as e:
            logger.error(f"Failed to initialize database schema: {e}")
            
    async def _initial_queue_discovery(self):
        """初始队列发现"""
        try:
            pattern = f"{self.prefix}:QUEUE:*"
            new_queues = set()
            
            async for key in self.redis_client.scan_iter(match=pattern, count=100):
                queue_name = key.decode('utf-8').split(":")[-1]
                new_queues.add(queue_name)
            
            if new_queues:
                all_queues = list(new_queues) + ['TASK_CHANGES']
                
                if self.consumer_manager:
                    self.consumer_manager.config['queues'] = all_queues
                    
                    if self.consumer_strategy == ConsumerStrategy.HEARTBEAT and hasattr(self.consumer_manager, '_heartbeat_strategy'):
                        actual_consumer_id = self.consumer_manager._heartbeat_strategy.consumer_id
                    else:
                        actual_consumer_id = self.consumer_id.rsplit('-', 1)[0] if '-' in self.consumer_id else self.consumer_id
                    
                    worker_key = f"{self.prefix}:{self.consumer_manager.config.get('worker_prefix', 'PG_CONSUMER')}:{actual_consumer_id}"
                    try:
                        self.consumer_manager.redis_client.hset(
                            worker_key, 
                            'queues', 
                            ','.join(all_queues)
                        )
                        logger.info(f"Initial queue discovery - found queues: {all_queues}")
                    except Exception as e:
                        logger.error(f"Error updating initial worker queues: {e}")
                
                self._known_queues = new_queues
                
        except Exception as e:
            logger.error(f"Error in initial queue discovery: {e}")
    
    async def _simulate_processing_delay(self, queue_name: str = None):
        """模拟处理延迟"""
        if self.test_mode == "slow_process" and self.processing_delay > 0:
            logger.info(f"[TEST] Simulating {self.processing_delay}s processing delay for {queue_name or 'message'}")
            await asyncio.sleep(self.processing_delay)
    
    async def _check_crash_condition(self, queue_name: str = None):
        """检查是否应该崩溃"""
        if self.test_mode == "crash_after_n":
            self.processed_count += 1
            if queue_name:
                self.queue_processed_count[queue_name] += 1
                count = self.queue_processed_count[queue_name]
            else:
                self.changes_processed_count += 1
                count = self.changes_processed_count
            
            logger.info(f"[TEST] Processed {count} messages from {queue_name or 'TASK_CHANGES'}")
            
            if self.crash_after_messages > 0 and count >= self.crash_after_messages:
                logger.error(f"[TEST] CRASHING after processing {count} messages from {queue_name or 'TASK_CHANGES'}!")
                # 强制退出进程
                os._exit(1)
    
    async def _consume_queue(self, queue_name: str):
        """消费单个队列的任务（带测试逻辑）"""
        stream_key = f"{self.prefix}:QUEUE:{queue_name}"
        check_backlog = True
        lastid = "0-0"
        
        consumer_name = self.consumer_id
        
        while self._running and queue_name in self._known_queues:
            try:
                myid = lastid if check_backlog else ">"
                
                # 减小批次大小便于测试
                count = 50 if self.test_mode != "normal" else 10000
                
                messages = await self.redis_client.xreadgroup(
                    self.consumer_group,
                    consumer_name,
                    {stream_key: myid},
                    count=count,
                    block=1000 if not check_backlog else 0
                )
                
                if not messages or (messages and len(messages[0][1]) == 0):
                    check_backlog = False
                    continue
                
                if messages:
                    # 模拟处理延迟
                    await self._simulate_processing_delay(queue_name)
                    
                    await self._process_messages(messages)
                    self._consecutive_errors[queue_name] = 0
                    
                    # 检查崩溃条件
                    await self._check_crash_condition(queue_name)
                    
                    if messages[0] and messages[0][1]:
                        lastid = messages[0][1][-1][0].decode('utf-8') if isinstance(messages[0][1][-1][0], bytes) else messages[0][1][-1][0]
                        check_backlog = len(messages[0][1]) >= count
                    
            except redis.ResponseError as e:
                if "NOGROUP" in str(e):
                    try:
                        await self.redis_client.xgroup_create(
                            stream_key, self.consumer_group, id='0', mkstream=True
                        )
                        logger.info(f"Recreated consumer group for queue: {queue_name}")
                        check_backlog = True
                        lastid = "0-0"
                    except:
                        pass
                else:
                    logger.error(f"Redis error for queue {queue_name}: {e}")
                    self._consecutive_errors[queue_name] += 1
                    
                if self._consecutive_errors[queue_name] > 10:
                    logger.warning(f"Too many errors for queue {queue_name}, will retry later")
                    await asyncio.sleep(30)
                    self._consecutive_errors[queue_name] = 0
                    
            except Exception as e:
                logger.error(f"Error consuming queue {queue_name}: {e}", exc_info=True)
                self._consecutive_errors[queue_name] += 1
                await asyncio.sleep(1)
    
    async def _consume_queues(self):
        """启动所有队列的消费任务"""
        queue_tasks = {}
        
        while self._running:
            try:
                for queue in self._known_queues:
                    if queue not in queue_tasks or queue_tasks[queue].done():
                        queue_tasks[queue] = asyncio.create_task(self._consume_queue(queue))
                        logger.info(f"Started consumer task for queue: {queue}")
                
                for queue in list(queue_tasks.keys()):
                    if queue not in self._known_queues:
                        queue_tasks[queue].cancel()
                        del queue_tasks[queue]
                        logger.info(f"Stopped consumer task for removed queue: {queue}")
                
                await asyncio.sleep(10)
                
            except Exception as e:
                logger.error(f"Error in consume_queues manager: {e}")
                await asyncio.sleep(5)
        
        for task in queue_tasks.values():
            task.cancel()
        
        await asyncio.gather(*queue_tasks.values(), return_exceptions=True)
                
    async def _process_messages(self, messages: List):
        """处理消息并保存到PostgreSQL（简化版）"""
        tasks_to_insert = []
        ack_batch = []
        
        for stream_key, stream_messages in messages:
            if not stream_messages:
                continue
                
            stream_key_str = stream_key.decode('utf-8') if isinstance(stream_key, bytes) else stream_key
            queue_name = stream_key_str.split(":")[-1]
            msg_ids_to_ack = []
            
            for msg_id, data in stream_messages:
                try:
                    if not msg_id or not data:
                        continue
                    
                    msg_id_str = msg_id.decode('utf-8') if isinstance(msg_id, bytes) else str(msg_id)
                    
                    logger.info(f"[TEST] Processing message {msg_id_str} from queue {queue_name}")
                    
                    # 简化处理，主要用于测试
                    task_info = {
                        'id': msg_id_str,
                        'queue_name': queue_name,
                        'task_name': 'test_task',
                        'task_data': '{}',
                        'priority': 0,
                        'retry_count': 0,
                        'max_retry': 3,
                        'status': 'pending',
                        'metadata': '{}',
                        'created_at': datetime.now(tz=timezone.utc)
                    }
                    tasks_to_insert.append(task_info)
                    msg_ids_to_ack.append(msg_id)
                    
                except Exception as e:
                    logger.error(f"Error processing message {msg_id}: {e}")
            
            if msg_ids_to_ack:
                ack_batch.append((stream_key, msg_ids_to_ack))
        
        if tasks_to_insert:
            # 简化：不真正插入数据库，只记录日志
            logger.info(f"[TEST] Would insert {len(tasks_to_insert)} tasks to PostgreSQL")
            
            if ack_batch:
                pipeline = self.redis_client.pipeline()
                for stream_key, msg_ids in ack_batch:
                    pipeline.xack(stream_key, self.consumer_group, *msg_ids)
                
                try:
                    await pipeline.execute()
                    total_acked = sum(len(msg_ids) for _, msg_ids in ack_batch)
                    logger.info(f"[TEST] Successfully ACKed {total_acked} messages")
                except Exception as e:
                    logger.error(f"Error executing batch ACK: {e}")
    
    async def _consume_task_changes(self):
        """消费任务变更事件流（支持pending消息恢复）"""
        change_stream_key = f"{self.prefix}:TASK_CHANGES"
        consumer_group = f"{self.prefix}_changes_consumer"
        
        consumer_name = self.consumer_manager.get_consumer_name('pg_consumer')
        
        # 创建消费者组
        try:
            await self.redis_client.xgroup_create(
                change_stream_key, consumer_group, id='0', mkstream=True
            )
            logger.info(f"Created consumer group for task changes stream")
        except redis.ResponseError:
            pass
        
        # 模仿 listen_event_by_task 的写法
        check_backlog = True
        lastid = "0-0"
        batch_size = 20 if self.test_mode != "normal" else 100
        
        while self._running:
            try:
                if check_backlog:
                    myid = lastid
                else:
                    myid = ">"
                
                messages = await self.redis_client.xreadgroup(
                    consumer_group,
                    consumer_name,
                    {change_stream_key: myid},
                    count=batch_size,
                    block=1000 if not check_backlog else 0
                )
                
                if not messages:
                    check_backlog = False
                    continue
                
                if messages and len(messages[0][1]) > 0:
                    check_backlog = len(messages[0][1]) >= batch_size
                else:
                    check_backlog = False
                
                task_ids_to_update = set()
                ack_ids = []
                
                for _, stream_messages in messages:
                    for msg_id, data in stream_messages:
                        try:
                            if isinstance(msg_id, bytes):
                                lastid = msg_id.decode('utf-8')
                            else:
                                lastid = str(msg_id)
                            
                            event_id = data.get(b'event_id')
                            if event_id:
                                if isinstance(event_id, bytes):
                                    event_id = event_id.decode('utf-8')
                                logger.info(f"[TEST] Processing TASK_CHANGES event: {event_id}")
                                task_ids_to_update.add(event_id)
                            ack_ids.append(msg_id)
                        except Exception as e:
                            logger.error(f"Error processing change event {msg_id}: {e}")
                
                if task_ids_to_update:
                    # 模拟处理延迟
                    await self._simulate_processing_delay("TASK_CHANGES")
                    
                    # 简化：不真正更新数据库
                    logger.info(f"[TEST] Would update {len(task_ids_to_update)} tasks from change events")
                    
                    # 检查崩溃条件
                    await self._check_crash_condition("TASK_CHANGES")
                
                if ack_ids:
                    await self.redis_client.xack(change_stream_key, consumer_group, *ack_ids)
                    logger.info(f"[TEST] ACKed {len(ack_ids)} TASK_CHANGES messages")
                
            except redis.ResponseError as e:
                if "NOGROUP" in str(e):
                    try:
                        await self.redis_client.xgroup_create(
                            change_stream_key, consumer_group, id='0', mkstream=True
                        )
                        logger.info(f"Recreated consumer group for task changes stream")
                        check_backlog = True
                        lastid = "0-0"
                    except:
                        pass
                else:
                    logger.error(f"Redis error in consume_task_changes: {e}")
                    await asyncio.sleep(1)
            except Exception as e:
                logger.error(f"Error in consume_task_changes: {e}", exc_info=True)
                await asyncio.sleep(1)
    
    async def _start_offline_recovery(self):
        """启动离线worker恢复服务"""
        logger.info("Starting offline worker recovery service for PG_CONSUMER")
        
        while self._running:
            try:
                total_recovered = 0
                
                # 恢复普通队列的消息
                for queue in self._known_queues:
                    try:
                        recovered = await self.offline_recovery.recover_offline_workers(
                            queue=queue,
                            current_consumer_name=self.consumer_id,
                            process_message_callback=self._process_recovered_message
                        )
                        
                        if recovered > 0:
                            logger.info(f"[TEST] Recovered {recovered} messages from queue {queue}")
                            total_recovered += recovered
                            
                    except Exception as e:
                        logger.error(f"Error recovering queue {queue}: {e}")
                
                if total_recovered > 0:
                    logger.info(f"[TEST] Total recovered {total_recovered} messages in this cycle")
                
                # 减少扫描频率便于测试
                await asyncio.sleep(5)
                
            except Exception as e:
                logger.error(f"Error in offline recovery service: {e}")
                await asyncio.sleep(10)
    
    async def _process_recovered_message(self, msg_id, msg_data, queue, consumer_id):
        """处理恢复的消息"""
        try:
            logger.info(f"[TEST] Processing recovered message {msg_id} from queue {queue}, offline worker {consumer_id}")
            
            # ACK消息
            if queue == 'TASK_CHANGES':
                stream_key = f"{self.prefix}:TASK_CHANGES"
                consumer_group = f"{self.prefix}_changes_consumer"
            else:
                stream_key = f"{self.prefix}:QUEUE:{queue}"
                consumer_group = self.consumer_group
                
            await self.redis_client.xack(stream_key, consumer_group, msg_id)
            logger.info(f"[TEST] ACKed recovered message {msg_id}")
            
        except Exception as e:
            logger.error(f"Error processing recovered message {msg_id}: {e}")


async def run_test_pg_consumer(pg_config: PostgreSQLConfig, redis_config: RedisConfig,
                               test_mode: str = "normal",
                               processing_delay: float = 0,
                               crash_after_messages: int = 0,
                               node_id: str = None):
    """运行测试版PostgreSQL消费者"""
    consumer = TestPostgreSQLConsumer(
        pg_config, 
        redis_config,
        node_id=node_id,
        consumer_strategy=ConsumerStrategy.HEARTBEAT,
        test_mode=test_mode,
        processing_delay=processing_delay,
        crash_after_messages=crash_after_messages
    )
    
    try:
        await consumer.start()
        while True:
            await asyncio.sleep(1)
            
    except KeyboardInterrupt:
        logger.info("Received interrupt signal")
    finally:
        await consumer.stop()


def main():
    """主入口函数"""
    import argparse
    from dotenv import load_dotenv
    
    load_dotenv()
    
    parser = argparse.ArgumentParser(description='Test PostgreSQL Consumer')
    parser.add_argument('--mode', choices=['normal', 'slow_process', 'crash_after_n'], 
                       default='normal', help='Test mode')
    parser.add_argument('--delay', type=float, default=0, 
                       help='Processing delay in seconds')
    parser.add_argument('--crash-after', type=int, default=0,
                       help='Crash after processing N messages')
    parser.add_argument('--node-id', type=str, default=None,
                       help='Node ID for this consumer')
    
    args = parser.parse_args()
    
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    
    pg_config = PostgreSQLConfig(
        host=os.getenv('JETTASK_PG_HOST', 'localhost'),
        port=int(os.getenv('JETTASK_PG_PORT', '5432')),
        database=os.getenv('JETTASK_PG_DB', 'jettask'),
        user=os.getenv('JETTASK_PG_USER', 'jettask'),
        password=os.getenv('JETTASK_PG_PASSWORD', '123456'),
    )
    
    redis_config = RedisConfig(
        host=os.getenv('REDIS_HOST', 'localhost'),
        port=int(os.getenv('REDIS_PORT', '6379')),
        db=int(os.getenv('REDIS_DB', '0')),
        password=os.getenv('REDIS_PASSWORD'),
    )
    
    logger.info(f"Starting test consumer with mode={args.mode}, delay={args.delay}, crash_after={args.crash_after}")
    
    asyncio.run(run_test_pg_consumer(
        pg_config, 
        redis_config,
        test_mode=args.mode,
        processing_delay=args.delay,
        crash_after_messages=args.crash_after,
        node_id=args.node_id
    ))


if __name__ == '__main__':
    main()


"""
错误处理工具
"""
import sys
import functools
from contextlib import contextmanager
from ..exceptions import TaskExecutionError


@contextmanager
def clean_task_errors():
    """
    上下文管理器，用于清理任务执行错误的堆栈
    
    使用示例:
        with clean_task_errors():
            result = await app.get_result(event_id, wait=True)
    """
    try:
        yield
    except TaskExecutionError as e:
        # 直接打印错误信息，不显示框架堆栈
        print(f"\n任务执行失败 (Task ID: {e.task_id}):", file=sys.stderr)
        print("-" * 60, file=sys.stderr)
        if e.error_traceback and e.error_traceback != "Task execution failed":
            print(e.error_traceback, file=sys.stderr)
        else:
            print("Task execution failed (no detailed error information available)", file=sys.stderr)
        print("-" * 60, file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        print("\n操作被用户中断", file=sys.stderr)
        sys.exit(130)
    except Exception as e:
        # 其他异常正常抛出
        raise


def handle_task_error(func):
    """
    装饰器，用于处理任务执行错误
    
    使用示例:
        @handle_task_error
        async def main():
            result = await app.get_result(event_id, wait=True)
    """
    @functools.wraps(func)
    def sync_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except TaskExecutionError as e:
            print(f"\n任务执行失败 (Task ID: {e.task_id}):", file=sys.stderr)
            print("-" * 60, file=sys.stderr)
            if e.error_traceback and e.error_traceback != "Task execution failed":
                print(e.error_traceback, file=sys.stderr)
            else:
                print("Task execution failed (no detailed error information available)", file=sys.stderr)
            print("-" * 60, file=sys.stderr)
            sys.exit(1)
    
    @functools.wraps(func)
    async def async_wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except TaskExecutionError as e:
            print(f"\n任务执行失败 (Task ID: {e.task_id}):", file=sys.stderr)
            print("-" * 60, file=sys.stderr)
            if e.error_traceback and e.error_traceback != "Task execution failed":
                print(e.error_traceback, file=sys.stderr)
            else:
                print("Task execution failed (no detailed error information available)", file=sys.stderr)
            print("-" * 60, file=sys.stderr)
            sys.exit(1)
    
    # 根据函数类型返回对应的包装器
    import asyncio
    if asyncio.iscoroutinefunction(func):
        return async_wrapper
    else:
        return sync_wrapper


"""
Task Router for modular task organization
Similar to FastAPI's APIRouter
"""

from typing import Optional, Dict, List, Any, Callable
from functools import wraps
import logging

logger = logging.getLogger(__name__)


class TaskRouter:
    """
    任务路由器，用于模块化组织任务
    
    使用示例：
        # tasks/email_tasks.py
        from jettask import TaskRouter
        
        router = TaskRouter(prefix="email", queue="emails")
        
        @router.task()
        async def send_welcome_email(user_id: str):
            # 实际任务名会是: email.send_welcome_email
            # 默认队列会是: emails
            pass
        
        # main.py
        from jettask import Jettask
        from tasks.email_tasks import router as email_router
        
        app = Jettask(redis_url="redis://localhost:6379/0")
        app.register_router(email_router)
    """
    
    def __init__(
        self,
        prefix: str = None,
        queue: str = None,
        tags: List[str] = None,
        default_timeout: int = None,
        default_max_retries: int = None,
        default_retry_delay: int = None,
    ):
        """
        初始化任务路由器
        
        Args:
            prefix: 任务名称前缀
            queue: 默认队列名
            tags: 标签列表，用于任务分组
            default_timeout: 默认超时时间
            default_max_retries: 默认最大重试次数
            default_retry_delay: 默认重试延迟
        """
        self.prefix = prefix
        self.default_queue = queue
        self.tags = tags or []
        self.default_timeout = default_timeout
        self.default_max_retries = default_max_retries
        self.default_retry_delay = default_retry_delay
        
        # 存储注册的任务
        self._tasks: Dict[str, Dict[str, Any]] = {}
        
    def task(
        self,
        name: str = None,
        queue: str = None,
        timeout: int = None,
        max_retries: int = None,
        retry_delay: int = None,
        **kwargs
    ):
        """
        任务装饰器
        
        Args:
            name: 任务名称（可选，默认使用函数名）
            queue: 队列名（可选，默认使用路由器的默认队列）
            timeout: 超时时间
            max_retries: 最大重试次数
            retry_delay: 重试延迟
            **kwargs: 其他任务参数
        """
        def decorator(func: Callable):
            # 生成任务名
            task_name = name or func.__name__
            if self.prefix:
                full_task_name = f"{self.prefix}.{task_name}"
            else:
                full_task_name = task_name
            
            # 合并参数（优先使用任务级别的参数）
            task_config = {
                'func': func,
                'name': full_task_name,
                'queue': queue or self.default_queue,
                'timeout': timeout or self.default_timeout,
                'max_retries': max_retries or self.default_max_retries,
                'retry_delay': retry_delay or self.default_retry_delay,
                'tags': self.tags,
                **kwargs
            }
            
            # 移除None值
            task_config = {k: v for k, v in task_config.items() if v is not None}
            
            # 存储任务配置
            self._tasks[full_task_name] = task_config
            
            # 返回原函数，保持函数可以被直接调用
            return func
            
        return decorator
    
    def include_router(self, router: 'TaskRouter', prefix: str = None):
        """
        包含另一个路由器（子路由器）
        
        Args:
            router: 要包含的路由器
            prefix: 额外的前缀
        """
        # 合并前缀
        if prefix:
            if self.prefix:
                combined_prefix = f"{self.prefix}.{prefix}"
            else:
                combined_prefix = prefix
        else:
            combined_prefix = self.prefix
        
        # 复制任务时更新前缀
        for task_name, task_config in router._tasks.items():
            # 更新任务名
            if combined_prefix and not task_name.startswith(combined_prefix):
                if router.prefix and task_name.startswith(router.prefix):
                    # 替换原有前缀
                    new_name = task_name.replace(router.prefix, combined_prefix, 1)
                else:
                    new_name = f"{combined_prefix}.{task_name}"
            else:
                new_name = task_name
            
            # 复制配置
            new_config = task_config.copy()
            new_config['name'] = new_name
            
            # 如果没有指定队列，使用当前路由器的默认队列
            if 'queue' not in new_config or new_config['queue'] is None:
                new_config['queue'] = self.default_queue
            
            self._tasks[new_name] = new_config
    
    def get_tasks(self) -> Dict[str, Dict[str, Any]]:
        """获取所有注册的任务"""
        return self._tasks


"""
Worker 扫描器 - 高效检测超时 Worker
使用 Redis Sorted Set 实现 O(log N) 复杂度
"""

import time
import logging
import asyncio
from typing import Dict, List, Optional

logger = logging.getLogger('app')


class WorkerScanner:
    """
    使用 Redis Sorted Set 优化的 Worker 扫描器
    
    核心优化：
    1. O(log N) 的超时检测复杂度
    2. 自动一致性维护
    3. 原子性操作保证数据一致
    """
    
    def __init__(self, sync_redis, async_redis, redis_prefix: str = 'jettask',
                 heartbeat_timeout: float = 3.0):
        self.redis = sync_redis
        self.async_redis = async_redis
        self.redis_prefix = redis_prefix
        self.heartbeat_timeout = heartbeat_timeout
        self.active_workers_key = f"{redis_prefix}:ACTIVE_WORKERS"
        
        # 一致性维护
        self._initialized = False
        self._last_full_sync = 0
        self._full_sync_interval = 60  # 每60秒完整同步
        self._scan_counter = 0
        self._partial_check_interval = 10  # 每10次扫描做部分检查
        
    async def ensure_initialized(self):
        """确保 Sorted Set 已初始化并保持一致"""
        current_time = time.time()
        
        # 首次初始化或定期完整同步
        if not self._initialized or (current_time - self._last_full_sync > self._full_sync_interval):
            await self._full_sync()
            self._initialized = True
            self._last_full_sync = current_time
            
    async def _full_sync(self):
        """完整同步 Hash 和 Sorted Set"""
        try:
            logger.debug("Starting full synchronization of ACTIVE_WORKERS")
            
            # 1. 收集所有活跃 worker 的 Hash 数据
            pattern = f"{self.redis_prefix}:WORKER:*"
            hash_workers = {}
            
            async for key in self.async_redis.scan_iter(match=pattern, count=100):
                if ':HISTORY:' not in key and ':REUSE:' not in key:
                    worker_id = key.split(':')[-1]
                    worker_data = await self.async_redis.hgetall(key)
                    
                    if worker_data and worker_data.get('is_alive', 'true').lower() == 'true':
                        try:
                            heartbeat = float(worker_data.get('last_heartbeat', 0))
                            hash_workers[worker_id] = heartbeat
                        except (ValueError, TypeError):
                            continue
            
            # 2. 获取 Sorted Set 中的数据
            zset_workers = await self.async_redis.zrange(
                self.active_workers_key, 0, -1, withscores=True
            )
            zset_dict = {worker_id: score for worker_id, score in zset_workers}
            
            # 3. 计算差异
            hash_ids = set(hash_workers.keys())
            zset_ids = set(zset_dict.keys())
            
            to_add = hash_ids - zset_ids  # Hash有但ZSet无
            to_remove = zset_ids - hash_ids  # ZSet有但Hash无
            to_update = {}  # 时间戳不一致的
            
            for worker_id in hash_ids & zset_ids:
                if abs(hash_workers[worker_id] - zset_dict[worker_id]) > 0.1:
                    to_update[worker_id] = hash_workers[worker_id]
            
            # 4. 批量修复
            if to_add or to_remove or to_update:
                pipeline = self.async_redis.pipeline()
                
                if to_add:
                    members = {worker_id: hash_workers[worker_id] for worker_id in to_add}
                    pipeline.zadd(self.active_workers_key, members)
                    
                if to_remove:
                    pipeline.zrem(self.active_workers_key, *to_remove)
                    
                if to_update:
                    pipeline.zadd(self.active_workers_key, to_update)
                    
                await pipeline.execute()
                logger.info(f"Full sync: +{len(to_add)}, -{len(to_remove)}, ~{len(to_update)}")
                
        except Exception as e:
            logger.error(f"Full sync failed: {e}")
    
    async def scan_timeout_workers(self) -> List[Dict]:
        """
        快速扫描超时的 worker - O(log N) 复杂度
        注意：需要考虑每个worker自己的heartbeat_timeout
        """
        await self.ensure_initialized()
        
        # 定期部分检查
        self._scan_counter += 1
        if self._scan_counter >= self._partial_check_interval:
            self._scan_counter = 0
            asyncio.create_task(self._partial_check())
        
        current_time = time.time()
        # 使用最大可能的超时时间作为初始筛选（避免遗漏）
        # 实际超时判断会在后面使用每个worker自己的timeout
        max_possible_timeout = 300  # 5分钟，足够覆盖大多数情况
        cutoff_time = current_time - max_possible_timeout
        
        # O(log N) 获取可能超时的 workers（宽松筛选）
        potential_timeout_worker_ids = await self.async_redis.zrangebyscore(
            self.active_workers_key, 
            min=0, 
            max=current_time - 1  # 至少1秒没更新的
        )
        
        if not potential_timeout_worker_ids:
            return []
        
        # 批量获取 worker 详细信息
        pipeline = self.async_redis.pipeline()
        for worker_id in potential_timeout_worker_ids:
            worker_key = f"{self.redis_prefix}:WORKER:{worker_id}"
            pipeline.hgetall(worker_key)
        
        workers_data = await pipeline.execute()
        
        # 验证并构造结果
        result = []
        cleanup_pipeline = self.async_redis.pipeline()
        need_cleanup = False
        
        for worker_id, worker_data in zip(potential_timeout_worker_ids, workers_data):
            if not worker_data:
                # Hash 不存在，清理 ZSet
                cleanup_pipeline.zrem(self.active_workers_key, worker_id)
                need_cleanup = True
                continue
            
            # 获取该worker自己的heartbeat_timeout
            worker_heartbeat_timeout = float(worker_data.get('heartbeat_timeout', self.heartbeat_timeout))
            
            # 双重检查实际心跳时间（使用worker自己的timeout）
            last_heartbeat = float(worker_data.get('last_heartbeat', 0))
            worker_cutoff_time = current_time - worker_heartbeat_timeout
            
            if last_heartbeat >= worker_cutoff_time:
                # 未超时，更新 ZSet 并跳过
                cleanup_pipeline.zadd(self.active_workers_key, {worker_id: last_heartbeat})
                need_cleanup = True
                continue
            
            # 检查存活状态
            is_alive = worker_data.get('is_alive', 'true').lower() == 'true'
            if not is_alive:
                # 已离线，清理 ZSet
                cleanup_pipeline.zrem(self.active_workers_key, worker_id)
                need_cleanup = True
                continue
            
            # 确认超时（使用worker自己的timeout判断）
            logger.debug(f"Worker {worker_id} timeout: last_heartbeat={last_heartbeat}, "
                        f"timeout={worker_heartbeat_timeout}s, cutoff={worker_cutoff_time}")
            worker_key = f"{self.redis_prefix}:WORKER:{worker_id}"
            result.append({
                'worker_key': worker_key,
                'worker_data': worker_data,
                'worker_id': worker_id
            })
        
        if need_cleanup:
            await cleanup_pipeline.execute()
        
        if result:
            logger.info(f"Found {len(result)} timeout workers")
        
        return result
    
    async def update_heartbeat(self, worker_id: str, heartbeat_time: Optional[float] = None):
        """原子性更新心跳（同时更新 Hash 和 ZSet）"""
        if heartbeat_time is None:
            heartbeat_time = time.time()
            
        pipeline = self.async_redis.pipeline()
        worker_key = f"{self.redis_prefix}:WORKER:{worker_id}"
        
        # 同时更新两个数据结构
        pipeline.hset(worker_key, 'last_heartbeat', str(heartbeat_time))
        pipeline.zadd(self.active_workers_key, {worker_id: heartbeat_time})
        
        await pipeline.execute()
    
    async def add_worker(self, worker_id: str, worker_data: Dict):
        """添加新 worker（同时加入 Hash 和 ZSet）"""
        heartbeat_time = float(worker_data.get('last_heartbeat', time.time()))
        
        pipeline = self.async_redis.pipeline()
        worker_key = f"{self.redis_prefix}:WORKER:{worker_id}"
        
        # 创建 Hash
        pipeline.hset(worker_key, mapping=worker_data)
        # 加入 ZSet
        pipeline.zadd(self.active_workers_key, {worker_id: heartbeat_time})
        
        await pipeline.execute()
        logger.debug(f"Added worker {worker_id} to system")
    
    async def remove_worker(self, worker_id: str):
        """移除 worker（同时更新 Hash 和 ZSet）"""
        pipeline = self.async_redis.pipeline()
        worker_key = f"{self.redis_prefix}:WORKER:{worker_id}"
        
        # 标记 Hash 为离线
        pipeline.hset(worker_key, 'is_alive', 'false')
        # 从 ZSet 移除
        pipeline.zrem(self.active_workers_key, worker_id)
        
        await pipeline.execute()
        logger.debug(f"Removed worker {worker_id} from active set")
    
    async def _partial_check(self):
        """部分一致性检查"""
        try:
            # 随机检查10个 worker
            sample_size = min(10, await self.async_redis.zcard(self.active_workers_key))
            if sample_size == 0:
                return
            
            # 随机抽样
            random_workers = await self.async_redis.zrandmember(
                self.active_workers_key, sample_size, withscores=True
            )
            
            for worker_id, zset_score in random_workers:
                worker_key = f"{self.redis_prefix}:WORKER:{worker_id}"
                hash_heartbeat = await self.async_redis.hget(worker_key, 'last_heartbeat')
                
                if not hash_heartbeat:
                    # Hash 不存在，删除 ZSet 条目
                    await self.async_redis.zrem(self.active_workers_key, worker_id)
                    logger.debug(f"Partial check: removed {worker_id}")
                else:
                    # 检查时间戳一致性
                    hash_time = float(hash_heartbeat)
                    if abs(hash_time - zset_score) > 1.0:
                        await self.async_redis.zadd(self.active_workers_key, {worker_id: hash_time})
                        logger.debug(f"Partial check: synced {worker_id}")
                        
        except Exception as e:
            logger.debug(f"Partial check error: {e}")
    
    async def get_active_count(self) -> int:
        """获取活跃 worker 数量 - O(1)"""
        return await self.async_redis.zcard(self.active_workers_key)


#!/usr/bin/env python3
"""
启动独立的JetTask Monitor后端API服务
"""
import uvicorn
import logging
import sys
import os

# 添加项目根目录到Python路径
current_dir = os.path.dirname(os.path.abspath(__file__))
project_root = os.path.dirname(os.path.dirname(os.path.dirname(current_dir)))
sys.path.insert(0, project_root)

# 设置日志
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

def main():
    """启动服务"""
    logger.info("启动JetTask Monitor独立后端API服务...")
    
    try:
        uvicorn.run(
            "main:app",
            host="0.0.0.0",
            port=8001,
            reload=True,  # 开发模式，生产环境应设为False
            log_level="info",
            access_log=True
        )
    except KeyboardInterrupt:
        logger.info("服务被用户中断")
    except Exception as e:
        logger.error(f"服务启动失败: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3
"""
JetTask WebUI 启动脚本
支持多种界面模式：
1. FastAPI + HTML (原始模式)
2. Gradio 界面 (新模式)
"""

import argparse
import asyncio
import logging
import os
import sys
from pathlib import Path

# 添加项目根目录到Python路径
project_root = Path(__file__).parent.parent.parent
sys.path.insert(0, str(project_root))

from jettask.webui.config import PostgreSQLConfig


def run_fastapi_server(host: str = "0.0.0.0", port: int = 8000, with_consumer: bool = False):
    """运行FastAPI服务器"""
    import uvicorn
    from jettask.webui.api import app
    
    # 如果需要启动消费者
    if with_consumer:
        pg_config = PostgreSQLConfig.from_env()
        if pg_config.dsn:
            app.state.pg_config = pg_config
            app.state.enable_consumer = True
            print("PostgreSQL consumer will be started with the server")
        else:
            print("Warning: PostgreSQL configuration not found, consumer disabled")
    
    print(f"Starting FastAPI server on http://{host}:{port}")
    print("Access the web interface at http://localhost:8000/")
    
    uvicorn.run(
        "jettask.webui.api:app",
        host=host,
        port=port,
        reload=False,
        log_level="info"
    )


def run_gradio_interface(host: str = "0.0.0.0", port: int = 7860, share: bool = False):
    """运行Gradio界面"""
    # 确保FastAPI服务器正在运行
    print("Note: Make sure the FastAPI server is running on port 8000")
    print("You can start it with: python -m jettask.webui.run_webui --mode fastapi")
    print()
    
    from jettask.webui.gradio_app import create_interface
    
    app = create_interface()
    print(f"Starting Gradio interface on http://{host}:{port}")
    
    app.launch(
        server_name=host,
        server_port=port,
        share=share,
        inbrowser=True
    )


def run_combined_mode(host: str = "0.0.0.0", api_port: int = 8000, ui_port: int = 7860):
    """同时运行FastAPI和Gradio"""
    import subprocess
    import time
    
    # 启动FastAPI服务器
    print("Starting FastAPI server...")
    api_process = subprocess.Popen([
        sys.executable, "-m", "jettask.webui.api"
    ])
    
    # 等待API服务器启动
    time.sleep(3)
    
    # 启动Gradio界面
    print("\nStarting Gradio interface...")
    try:
        run_gradio_interface(host, ui_port, share=False)
    except KeyboardInterrupt:
        print("\nShutting down...")
    finally:
        api_process.terminate()
        api_process.wait()


def run_integrated_gradio(host: str = "0.0.0.0", port: int = 7860, share: bool = False):
    """运行集成的Gradio界面（直接访问数据源）"""
    from jettask.webui.integrated_gradio_app import create_integrated_interface
    
    print("Starting Integrated Gradio interface (direct data access)...")
    print(f"Access the interface at http://{host}:{port}")
    
    app = create_integrated_interface()
    app.launch(
        server_name=host,
        server_port=port,
        share=share,
        inbrowser=True
    )


def main():
    parser = argparse.ArgumentParser(description="JetTask WebUI 启动器")
    parser.add_argument(
        "--mode", 
        choices=["fastapi", "gradio", "combined", "integrated"],
        default="integrated",
        help="界面模式: fastapi(原始HTML), gradio(新界面), combined(同时启动), integrated(集成版,直接访问数据)"
    )
    parser.add_argument("--host", default="0.0.0.0", help="服务器地址")
    parser.add_argument("--port", type=int, help="服务器端口")
    parser.add_argument("--with-consumer", action="store_true", help="启动PostgreSQL消费者")
    parser.add_argument("--share", action="store_true", help="创建Gradio公共链接")
    
    args = parser.parse_args()
    
    # 设置日志
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    
    if args.mode == "fastapi":
        port = args.port or 8000
        run_fastapi_server(args.host, port, args.with_consumer)
    elif args.mode == "gradio":
        port = args.port or 7860
        run_gradio_interface(args.host, port, args.share)
    elif args.mode == "combined":
        run_combined_mode(args.host)
    elif args.mode == "integrated":
        port = args.port or 7860
        run_integrated_gradio(args.host, port, args.share)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()


import multiprocessing
import logging
import time
import os
import signal
from typing import Dict, List
from multiprocessing import Process, Event
from dataclasses import dataclass

from .base import BaseExecutor

logger = logging.getLogger('app')


@dataclass
class ProcessConfig:
    """Configuration for AsyncIO executor process"""
    executor_id: int
    redis_url: str
    queues: List[str]
    app_tasks: Dict
    consumer_strategy: str
    consumer_config: Dict
    max_connections: int
    prefetch_multiplier: int
    concurrency_per_process: int = 10000


class MultiAsyncioExecutor(BaseExecutor):
    """
    Multi-asyncio executor that manages multiple AsyncioExecutor instances
    Each instance runs in its own process with its own event loop
    
    Features:
    - Automatic process restart on failure
    - Graceful shutdown with timeout
    - Process health monitoring
    - Configurable concurrency per process
    """
    
    def __init__(self, event_queue, app, concurrency=3):
        super().__init__(event_queue, app, concurrency)
        self.processes: Dict[int, Process] = {}
        self.process_configs: Dict[int, ProcessConfig] = {}
        self.shutdown_event = Event()
        self._monitor_interval = 1  # seconds
        self._status_log_interval = 30  # seconds
        self._restart_delay = 2  # seconds
        self._max_restart_attempts = 3
        self._restart_counts: Dict[int, int] = {}
        self._main_received_signal = False  # Track if main process received signal
        
    def logic(self):
        """
        Logic method for BaseExecutor interface.
        Not used in MultiAsyncioExecutor as it delegates to AsyncioExecutor instances.
        """
        pass
        
    @staticmethod
    def _run_asyncio_executor(config: ProcessConfig, shutdown_event):
        """Run an AsyncioExecutor in a separate process"""
        # Set process name for debugging
        multiprocessing.current_process().name = f"AsyncioExecutor-{config.executor_id}"
        
        # Configure logging for subprocess
        logging.basicConfig(
            level=logging.INFO,
            format=f"%(asctime)s - %(levelname)s - [Executor-{config.executor_id}] - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        logger = logging.getLogger('app')
        
        # Handle signals gracefully
        def signal_handler(signum, _frame):
            logger.info(f"AsyncioExecutor #{config.executor_id} received signal {signum}")
            shutdown_event.set()
            
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        
        try:
            # Import inside process to avoid pickle issues
            from ..core.app import Jettask
            
            # Create app instance for this process
            app = Jettask(
                redis_url=config.redis_url,
                max_connections=config.max_connections,
                consumer_strategy=config.consumer_strategy,
                consumer_config=config.consumer_config,
                tasks=config.app_tasks
            )
            
            logger.info(f"Started AsyncioExecutor #{config.executor_id} in process {os.getpid()}")
            
            # Check shutdown event before starting
            if shutdown_event.is_set():
                logger.info(f"AsyncioExecutor #{config.executor_id} shutdown before start")
                return
            
            # Start the executor
            app._start(
                execute_type="asyncio",
                queues=config.queues,
                concurrency=config.concurrency_per_process,
                prefetch_multiplier=config.prefetch_multiplier
            )
            
        except KeyboardInterrupt:
            logger.info(f"AsyncioExecutor #{config.executor_id} received interrupt")
        except Exception as e:
            logger.error(f"AsyncioExecutor #{config.executor_id} error: {e}")
            import traceback
            traceback.print_exc()
            raise  # Re-raise to trigger restart mechanism
        finally:
            logger.info(f"AsyncioExecutor #{config.executor_id} stopped")
    
    def _create_process_config(self, executor_id: int, queues: List[str], 
                              prefetch_multiplier: int) -> ProcessConfig:
        """Create configuration for a process"""
        return ProcessConfig(
            executor_id=executor_id,
            redis_url=self.app.redis_url,
            queues=queues,
            app_tasks=self.app._tasks,
            consumer_strategy=self.app.consumer_strategy,
            consumer_config=self.app.consumer_config,
            max_connections=self.app.max_connections,
            prefetch_multiplier=prefetch_multiplier,
            concurrency_per_process=10000
        )
    
    def _start_process(self, executor_id: int, config: ProcessConfig) -> Process:
        """Start a single AsyncioExecutor process"""
        process = Process(
            target=self._run_asyncio_executor,
            args=(config, self.shutdown_event),
            name=f"AsyncioExecutor-{executor_id}"
        )
        process.start()
        logger.info(f"Started AsyncioExecutor process #{executor_id} (PID: {process.pid})")
        return process
    
    def _restart_process(self, executor_id: int):
        """Restart a failed process with exponential backoff"""
        if self.shutdown_event.is_set():
            return
            
        restart_count = self._restart_counts.get(executor_id, 0)
        if restart_count >= self._max_restart_attempts:
            logger.error(f"Process #{executor_id} exceeded max restart attempts ({self._max_restart_attempts})")
            return
            
        self._restart_counts[executor_id] = restart_count + 1
        delay = self._restart_delay * (2 ** restart_count)  # Exponential backoff
        
        logger.info(f"Restarting process #{executor_id} (attempt {restart_count + 1}/{self._max_restart_attempts}) "
                   f"after {delay}s delay")
        time.sleep(delay)
        
        config = self.process_configs[executor_id]
        process = self._start_process(executor_id, config)
        self.processes[executor_id] = process
    
    def _monitor_processes(self):
        """Monitor process health and restart failed processes"""
        alive_count = 0
        for executor_id, process in list(self.processes.items()):
            if process.is_alive():
                alive_count += 1
                # Reset restart count for healthy processes
                self._restart_counts[executor_id] = 0
            else:
                exit_code = process.exitcode
                # 如果正在关闭，不要重启进程
                if self.shutdown_event.is_set():
                    logger.info(f"Process #{executor_id} stopped during shutdown")
                # 如果进程收到SIGTERM（exit_code == -15）或SIGINT（exit_code == -2），认为是正常关闭
                elif exit_code == -15 or exit_code == -2:
                    logger.info(f"Process #{executor_id} received termination signal, marking as shutdown")
                    # 设置shutdown事件，避免重启其他进程
                    self.shutdown_event.set()
                # 如果所有进程都同时停止（正常退出），认为是收到了关闭信号
                elif exit_code == 0 or exit_code is None:
                    # 如果主进程已经收到信号，不要重启
                    if self._main_received_signal:
                        logger.info(f"Process #{executor_id} stopped after main received signal")
                    else:
                        # 检查是否所有进程都停止了
                        all_stopped = True
                        for _, p in self.processes.items():
                            if p.is_alive():
                                all_stopped = False
                                break
                        
                        if all_stopped:
                            logger.info(f"All processes stopped simultaneously, marking as shutdown")
                            self.shutdown_event.set()
                        else:
                            logger.warning(f"Process #{executor_id} stopped unexpectedly with exit code {exit_code}")
                            self._restart_process(executor_id)
                else:
                    logger.warning(f"Process #{executor_id} exited with code {exit_code}")
                    self._restart_process(executor_id)
        
        return alive_count
    
    def loop(self):
        """Main loop that starts and monitors AsyncioExecutor processes"""
        logger.info(f"Starting MultiAsyncioExecutor with {self.concurrency} asyncio processes")
        
        # Set up signal handler to track when main process receives signal
        def signal_handler(signum, frame):
            logger.info(f"MultiAsyncioExecutor received signal {signum}")
            self._main_received_signal = True
            self.shutdown_event.set()
        
        # Register signal handlers
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        
        try:
            # Get configuration
            queues = getattr(self.app.ep, 'queues', ['robust_bench'])
            prefetch_multiplier = getattr(self, 'prefetch_multiplier', 100)
            
            # Start AsyncioExecutor processes
            for i in range(self.concurrency):
                config = self._create_process_config(i, queues, prefetch_multiplier)
                self.process_configs[i] = config
                
                process = self._start_process(i, config)
                self.processes[i] = process
                
                # Small delay to avoid thundering herd
                time.sleep(0.1)
            
            logger.info(f"All {self.concurrency} AsyncioExecutor processes started")
            
            # Monitor executor processes
            last_status_log = time.time()
            
            while not self.shutdown_event.is_set():
                # Monitor process health
                alive_count = self._monitor_processes()
                
                if alive_count == 0:
                    if self._main_received_signal or self.shutdown_event.is_set():
                        logger.info("All AsyncioExecutor processes have stopped during shutdown")
                    else:
                        logger.error("All AsyncioExecutor processes have stopped unexpectedly")
                    break
                
                # Log status periodically
                current_time = time.time()
                if current_time - last_status_log > self._status_log_interval:
                    dead_count = self.concurrency - alive_count
                    logger.info(f"MultiAsyncioExecutor status - Active: {alive_count}/{self.concurrency}, "
                              f"Dead: {dead_count}, Restart attempts: {sum(self._restart_counts.values())}")
                    last_status_log = current_time
                
                time.sleep(self._monitor_interval)
                
        except KeyboardInterrupt:
            logger.info("MultiAsyncioExecutor received KeyboardInterrupt")
            self._main_received_signal = True
            self.shutdown_event.set()
        except Exception as e:
            logger.error(f"MultiAsyncioExecutor error: {e}")
            import traceback
            traceback.print_exc()
        finally:
            self.shutdown()
    
    def shutdown(self):
        """Gracefully shutdown all executor processes"""
        logger.info("Shutting down MultiAsyncioExecutor...")
        
        # Signal shutdown - this prevents any restart attempts
        self.shutdown_event.set()
        
        # 快速关闭模式 - 减少等待时间
        shutdown_timeout = 1.0  # 减少到1秒
        terminate_timeout = 0.5  # 减少到0.5秒
        
        # 先发送终止信号给所有进程
        for executor_id, process in self.processes.items():
            if process.is_alive():
                logger.info(f"Sending TERM signal to process #{executor_id} (PID: {process.pid})")
                process.terminate()
        
        # 短暂等待所有进程自行退出
        time.sleep(shutdown_timeout)
        
        # 强制杀死仍在运行的进程
        for executor_id, process in self.processes.items():
            if process.is_alive():
                logger.warning(f"Process #{executor_id} did not terminate, killing...")
                process.kill()
                # 不再等待join，让操作系统清理
        
        # Clear process tracking
        self.processes.clear()
        self.process_configs.clear()
        self._restart_counts.clear()
        
        logger.info("MultiAsyncioExecutor shutdown complete")


#!/usr/bin/env python
"""
独立的延迟队列扫描器进程
用于扫描延迟任务并投递到执行队列
"""

import asyncio
import time
import logging
import signal
import sys
from typing import Optional
import uuid

import redis.asyncio as aioredis
from redis.asyncio.lock import Lock

logger = logging.getLogger(__name__)


class DelayQueueScanner:
    """
    独立的延迟队列扫描器
    使用分布式锁确保只有一个扫描器实例在运行
    """
    
    def __init__(
        self,
        redis_url: str = "redis://localhost:6379/0",
        redis_prefix: str = "jettask",
        scan_interval: float = 1.0,
        batch_size: int = 100,
        lock_ttl: int = 10
    ):
        """
        初始化扫描器
        
        Args:
            redis_url: Redis连接URL
            redis_prefix: Redis键前缀
            scan_interval: 扫描间隔（秒）
            batch_size: 每次处理的任务数
            lock_ttl: 分布式锁TTL（秒）
        """
        self.redis_url = redis_url
        self.redis_prefix = redis_prefix
        self.scan_interval = scan_interval
        self.batch_size = batch_size
        self.lock_ttl = lock_ttl
        
        self.redis: Optional[aioredis.Redis] = None
        self.running = False
        self.scanner_id = f"scanner-{uuid.uuid4().hex[:8]}"
        
        # 设置信号处理
        signal.signal(signal.SIGINT, self._signal_handler)
        signal.signal(signal.SIGTERM, self._signal_handler)
    
    def _signal_handler(self, signum, frame):
        """信号处理器"""
        logger.info(f"Received signal {signum}, stopping scanner...")
        self.running = False
    
    async def connect(self):
        """建立Redis连接"""
        if not self.redis:
            self.redis = await aioredis.from_url(
                self.redis_url,
                encoding="utf-8",
                decode_responses=False
            )
            logger.info(f"Scanner {self.scanner_id} connected to Redis")
    
    async def disconnect(self):
        """关闭Redis连接"""
        if self.redis:
            await self.redis.close()
            self.redis = None
    
    def _get_queue_key(self) -> str:
        """获取延迟队列键名"""
        return f"{self.redis_prefix}:DELAYED:QUEUE"
    
    def _get_task_key(self, task_id: str) -> str:
        """获取任务详情键名"""
        return f"{self.redis_prefix}:DELAYED:TASK:{task_id}"
    
    def _get_lock_key(self) -> str:
        """获取分布式锁键名"""
        return f"{self.redis_prefix}:DELAYED:SCANNER_LOCK"
    
    def get_lock(self) -> Lock:
        """
        获取Redis分布式锁对象
        
        Returns:
            Redis Lock对象
        """
        return Lock(
            self.redis, 
            self._get_lock_key(),
            timeout=self.lock_ttl,  # 锁的自动过期时间
            blocking=True,  # 阻塞获取
            blocking_timeout=0.1  # 最多等待0.1秒
        )
    
    async def scan_and_deliver(self) -> int:
        """
        扫描并投递到期任务
        
        Returns:
            处理的任务数
        """
        # 使用Redis原生锁
        lock = self.get_lock()
        
        # 尝试获取锁
        async with lock:
            now = time.time()
            processed_count = 0
            
            # 使用ZPOPMIN原子地获取并移除到期任务
            for _ in range(self.batch_size):
                # 获取分数最小的任务
                result = await self.redis.zpopmin(self._get_queue_key(), 1)
                
                if not result:
                    break
                
                for item in result:
                    if len(item) == 2:
                        task_id_bytes, score = item
                        task_id = task_id_bytes.decode() if isinstance(task_id_bytes, bytes) else task_id_bytes
                        
                        # 检查是否到期
                        if score > now:
                            # 还没到期，放回队列
                            await self.redis.zadd(self._get_queue_key(), {task_id: score})
                            break  # 后面的任务肯定也没到期
                        
                        # 处理到期任务
                        try:
                            await self.deliver_task(task_id)
                            processed_count += 1
                        except Exception as e:
                            logger.error(f"Failed to deliver task {task_id}: {e}")
                            # 失败的任务延迟10秒重试
                            retry_time = time.time() + 10
                            await self.redis.zadd(self._get_queue_key(), {task_id: retry_time})
            
            if processed_count > 0:
                logger.info(f"Scanner {self.scanner_id} delivered {processed_count} tasks")
            
            return processed_count
    
    async def deliver_task(self, task_id: str):
        """
        投递任务到执行队列
        
        Args:
            task_id: 任务ID
        """
        # 获取任务详情
        task_json = await self.redis.get(self._get_task_key(task_id))
        
        if not task_json:
            logger.warning(f"Task {task_id} not found, skipping")
            return
        
        import json
        task_info = json.loads(task_json)
        
        queue_name = task_info['queue_name']
        task_data = task_info['task_data']
        
        # 写入到执行队列（Redis Stream）
        stream_key = f"{self.redis_prefix}:{queue_name}"
        await self.redis.xadd(stream_key, {'data': json.dumps(task_data)})
        
        # 更新任务状态
        event_id = task_data.get('event_id', task_id)
        status_key = f"{self.redis_prefix}:TASK:{event_id}"
        await self.redis.hset(status_key, "status", "pending")
        
        # 删除任务详情
        await self.redis.delete(self._get_task_key(task_id))
        
        logger.info(f"Delivered task {task_id} to queue {queue_name}")
    
    async def run(self):
        """运行扫描器"""
        await self.connect()
        
        self.running = True
        logger.info(f"Delay queue scanner {self.scanner_id} started")
        logger.info(f"Scan interval: {self.scan_interval}s, Batch size: {self.batch_size}")
        
        try:
            while self.running:
                try:
                    # 扫描并投递任务
                    await self.scan_and_deliver()
                except Exception as e:
                    logger.error(f"Scanner error: {e}", exc_info=True)
                
                # 等待下一次扫描
                await asyncio.sleep(self.scan_interval)
                
        finally:
            logger.info(f"Scanner {self.scanner_id} stopped")
            await self.disconnect()


async def main():
    """主函数"""
    import argparse
    
    # 配置日志
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    
    # 解析命令行参数
    parser = argparse.ArgumentParser(description='Jettask Delay Queue Scanner')
    parser.add_argument('--redis-url', default='redis://localhost:6379/0',
                        help='Redis connection URL')
    parser.add_argument('--redis-prefix', default='jettask',
                        help='Redis key prefix')
    parser.add_argument('--scan-interval', type=float, default=1.0,
                        help='Scan interval in seconds')
    parser.add_argument('--batch-size', type=int, default=100,
                        help='Number of tasks to process per scan')
    parser.add_argument('--lock-ttl', type=int, default=10,
                        help='Distributed lock TTL in seconds')
    
    args = parser.parse_args()
    
    # 创建并运行扫描器
    scanner = DelayQueueScanner(
        redis_url=args.redis_url,
        redis_prefix=args.redis_prefix,
        scan_interval=args.scan_interval,
        batch_size=args.batch_size,
        lock_ttl=args.lock_ttl
    )
    
    try:
        await scanner.run()
    except KeyboardInterrupt:
        logger.info("Scanner interrupted by user")


if __name__ == '__main__':
    asyncio.run(main())